#!/usr/bin/env python
#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Test the RegisterOutputData class."""

from __future__ import absolute_import
import sys
import unittest
from mock import patch, MagicMock as Mock

from Tests.Utilities.GeneralUtils import assertEqualsImproved, assertDiracFailsWith, \
    assertMockCalls, assertDiracSucceedsWith
from ILCDIRAC.Workflow.Modules.RegisterOutputData import RegisterOutputData
from DIRAC import S_OK, S_ERROR

__RCSID__ = "$Id$"

MODULE_NAME = 'ILCDIRAC.Workflow.Modules.RegisterOutputData'

# pylint: disable=too-many-public-methods


class TestRegisterOutputData(unittest.TestCase):
  """Test the different methods of the class."""

  def setUp(self):
    self.ops_mock = Mock()
    self.fcc_mock = Mock()
    mocked_modules = {'DIRAC.ConfigurationSystem.Client.Helpers.Operations': self.ops_mock,
                       'DIRAC.Resources.Catalog.FileCatalogClient': self.fcc_mock}
    self.module_patcher = patch.dict(sys.modules, mocked_modules)
    self.module_patcher.start()
    self.rod = RegisterOutputData()

  def tearDown(self):
    self.module_patcher.stop()

  def test_applicationspecificinputs(self):
    self.rod.step_commons['Enable'] = True
    self.rod.workflow_commons['ProductionOutputData'] = 'MyOutputLFN1.php;other__lfn.stdio;;last_file.stdhep'
    self.rod.workflow_commons['Luminosity'] = 1391.2
    self.rod.workflow_commons['Info'] = {'stdhepcut': {'Reduction': 12.2, 'CutEfficiency': 0.92}}
    assertDiracSucceedsWith(self.rod.applicationSpecificInputs(), 'Parameters resolved', self)
    assertEqualsImproved((self.rod.enable, self.rod.prodOutputLFNs, self.rod.luminosity,
                            self.rod.sel_eff, self.rod.cut_eff, self.rod.add_info),
                          (True, ['MyOutputLFN1.php', 'other__lfn.stdio', '', 'last_file.stdhep'],
                            1391.2, 12.2, 0.92, 'de'), self)
    assert 'stdhepcut' not in self.rod.workflow_commons['Info']

  def test_applicationspecificinputs_nonset(self):
    self.rod.workflow_commons['Info'] = {'SomeKey': 'someValue'}
    assertDiracSucceedsWith(self.rod.applicationSpecificInputs(), 'Parameters resolved', self)
    assertEqualsImproved((self.rod.enable, self.rod.prodOutputLFNs, self.rod.luminosity,
                            self.rod.nbofevents, self.rod.sel_eff, self.rod.cut_eff, self.rod.add_info),
                          (True, [], 0, 0, 0, 0, 'ds7:SomeKeys9:someValuee'), self)

  def test_applicationspecificinputs_othercases(self):
    self.rod.step_commons['Enable'] = {'mydict': True}
    assertDiracSucceedsWith(self.rod.applicationSpecificInputs(), 'Parameters resolved', self)
    assertEqualsImproved((self.rod.enable, self.rod.sel_eff, self.rod.cut_eff, self.rod.add_info),
                          (False, 0, 0, ''), self)

  def test_execute_resolveinput_fails(self):
    with patch('%s.RegisterOutputData.resolveInputVariables' % MODULE_NAME, new=Mock(return_value=S_ERROR('mock_var_testerr'))):
      assertDiracFailsWith(self.rod.execute(), 'mock_var_testerr', self)

  def test_execute_status_not_ok(self):
    self.rod.workflowStatus['OK'] = True
    self.rod.stepStatus['OK'] = False
    assertDiracSucceedsWith(self.rod.execute(), 'No registration of output data metadata attempted', self)

  def test_execute_no_production_output_data(self):
    assertDiracSucceedsWith(self.rod.execute(), "No files' metadata to be registered", self)

  def test_execute_minimal(self):
    self.rod.enable = False
    self.rod.prodOutputLFNs = ['myOutput.lfn.stdhep']
    with patch('%s.RegisterOutputData.resolveInputVariables' % MODULE_NAME, new=Mock(return_value=S_OK())):
      assertDiracSucceedsWith(self.rod.execute(), 'Output data metadata registered in catalog', self)

  def test_execute_maximal(self):
    self.rod.nbofevents = 1389
    self.rod.luminosity = 9814.2
    self.rod.sel_eff = 184.2
    self.rod.cut_eff = 13.1
    self.rod.InputData = 'myTestInputFiles.rec'
    self.rod.inputdataMeta['CrossSection'] = 'myTestCrosssection'
    self.rod.WorkflowStartFrom = 'EventZerotest'
    self.rod.add_info = 'MoreInfo.additional_testme'
    self.rod.workflow_commons['file_number_of_event_relation'] = {'myOutput.lfn.stdhep': 2148}
    self.rod.prodOutputLFNs = ['/some/dir/myOutput.lfn.stdhep', 'some_other.file', '/other/dir/lastOne', '']
    fcc_mock = Mock()
    fcc_mock.setMetadata.return_value = S_OK()
    fcc_mock.addFileAncestors.return_value = S_OK()
    self.rod.filecatalog = fcc_mock
    with patch('%s.RegisterOutputData.resolveInputVariables' % MODULE_NAME, new=Mock(return_value=S_OK())):
      assertDiracSucceedsWith(self.rod.execute(), 'Output data metadata registered in catalog', self)
      assertMockCalls(fcc_mock.setMetadata,
                      [('/some/dir/myOutput.lfn.stdhep',
                        {'NumberOfEvents': 2148,
                         'Luminosity': 9814.2,
                         'Reduction': 184.2,
                         'CutEfficiency': 13.1,
                         'AdditionalInfo': 'MoreInfo.additional_testme',
                         'CrossSection': 'myTestCrosssection',
                         'FirstEventFromInput': 'EventZerotest'}),
                          ('some_other.file',
                           {'NumberOfEvents': 1389,
                            'Luminosity': 9814.2,
                            'Reduction': 184.2,
                            'CutEfficiency': 13.1,
                            'AdditionalInfo': 'MoreInfo.additional_testme',
                            'CrossSection': 'myTestCrosssection',
                            'FirstEventFromInput': 'EventZerotest'}),
                          ('/other/dir/lastOne',
                           {'NumberOfEvents': 1389,
                            'Luminosity': 9814.2,
                            'Reduction': 184.2,
                            'CutEfficiency': 13.1,
                            'AdditionalInfo': 'MoreInfo.additional_testme',
                            'CrossSection': 'myTestCrosssection',
                            'FirstEventFromInput': 'EventZerotest'}),
                          ('',
                           {'NumberOfEvents': 1389,
                            'Luminosity': 9814.2,
                            'Reduction': 184.2,
                            'CutEfficiency': 13.1,
                            'AdditionalInfo': 'MoreInfo.additional_testme',
                            'CrossSection': 'myTestCrosssection',
                            'FirstEventFromInput': 'EventZerotest'})],
                      self)
      assertMockCalls(fcc_mock.addFileAncestors, [
          {'/some/dir/myOutput.lfn.stdhep': {'Ancestors': 'myTestInputFiles.rec'}},
          {'some_other.file': {'Ancestors': 'myTestInputFiles.rec'}},
          {'/other/dir/lastOne': {'Ancestors': 'myTestInputFiles.rec'}},
          {'': {'Ancestors': 'myTestInputFiles.rec'}}], self)

  def test_execute_maximal_othercase(self):
    self.rod.nbofevents = 1389
    self.rod.luminosity = 9814.2
    self.rod.sel_eff = 184.2
    self.rod.cut_eff = 13.1
    self.rod.InputData = 'myTestInputFiles.rec'
    self.rod.inputdataMeta['CrossSection'] = 'myTestCrosssection'
    self.rod.inputdataMeta['AdditionalInfo'] = 'more_information_testme'
    self.rod.WorkflowStartFrom = 'EventZerotest'
    self.rod.prodOutputLFNs = ['/some/test/dir/mytestfile.txt']
    fcc_mock = Mock()
    fcc_mock.setMetadata.return_value = S_OK()
    fcc_mock.addFileAncestors.return_value = S_OK()
    self.rod.filecatalog = fcc_mock
    with patch('%s.RegisterOutputData.resolveInputVariables' % MODULE_NAME, new=Mock(return_value=S_OK())):
      assertDiracSucceedsWith(self.rod.execute(), 'Output data metadata registered in catalog', self)
      fcc_mock.setMetadata.assert_called_once_with('/some/test/dir/mytestfile.txt', {
          'NumberOfEvents': 1389, 'Luminosity': 9814.2, 'Reduction': 184.2,
          'CutEfficiency': 13.1, 'AdditionalInfo': 'more_information_testme',
          'CrossSection': 'myTestCrosssection', 'FirstEventFromInput': 'EventZerotest'})
      fcc_mock.addFileAncestors.assert_called_once_with(
          {'/some/test/dir/mytestfile.txt': {'Ancestors': 'myTestInputFiles.rec'}})

  def test_execute_setmeta_fails(self):
    self.rod.nbofevents = 1389
    self.rod.luminosity = 9814.2
    self.rod.sel_eff = 184.2
    self.rod.cut_eff = 13.1
    self.rod.InputData = 'myTestInputFiles.rec'
    self.rod.inputdataMeta['CrossSection'] = 'myTestCrosssection'
    self.rod.inputdataMeta['AdditionalInfo'] = 'more_information_testme'
    self.rod.WorkflowStartFrom = 'EventZerotest'
    self.rod.prodOutputLFNs = ['/some/test/dir/mytestfile.txt']
    fcc_mock = Mock()
    fcc_mock.setMetadata.return_value = S_ERROR('test_err_metadata')
    fcc_mock.addFileAncestors.return_value = S_OK()
    self.rod.filecatalog = fcc_mock
    with patch('%s.RegisterOutputData.resolveInputVariables' % MODULE_NAME, new=Mock(return_value=S_OK())):
      assertDiracFailsWith(self.rod.execute(), 'test_err_metadata', self)

  def test_execute_maximal_addancestors_fails(self):
    self.rod.nbofevents = 1389
    self.rod.luminosity = 9814.2
    self.rod.sel_eff = 184.2
    self.rod.cut_eff = 13.1
    self.rod.InputData = 'myTestInputFiles.rec'
    self.rod.inputdataMeta['CrossSection'] = 'myTestCrosssection'
    self.rod.inputdataMeta['AdditionalInfo'] = 'more_information_testme'
    self.rod.WorkflowStartFrom = 'EventZerotest'
    self.rod.prodOutputLFNs = ['/some/test/dir/mytestfile.txt']
    fcc_mock = Mock()
    fcc_mock.setMetadata.return_value = S_OK()
    fcc_mock.addFileAncestors.return_value = S_ERROR('testme_addancestors_Err')
    self.rod.filecatalog = fcc_mock
    with patch('%s.RegisterOutputData.resolveInputVariables' % MODULE_NAME, new=Mock(return_value=S_OK())):
      assertDiracFailsWith(self.rod.execute(), 'testme_addancestors_Err', self)



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""tests for PrepareOptionFiles."""

from __future__ import print_function
from __future__ import absolute_import
import os
import shutil
import tempfile
import unittest
from xml.etree.ElementTree import ElementTree
from mock import patch, MagicMock as Mock

from Tests.Utilities.GeneralUtils import assertDiracFailsWith

__RCSID__ = "$Id$"

MODULE_NAME = 'ILCDIRAC.Core.Utilities.PrepareOptionFiles'


def cleanup(tempdir):
  """Remove files after run."""
  try:
    shutil.rmtree(tempdir)
  except OSError:
    pass


class TestPrepareMarlinXMLFileBase(unittest.TestCase):
  """tests for the perpareXMLFile function used to modify marlin steering files."""

  def setUp(self):
    self.basedir = os.getcwd()
    self.tmpdir = tempfile.mkdtemp("", dir="./")
    os.chdir(self.tmpdir)
    self.inputXMLFile = "marlininput.xml"
    self.testedTree = None
    self.expectedTree = None
    self.expectedOutput = ""

  def tearDown(self):
    """Remove the fake files."""
    os.chdir(self.basedir)
    cleanup(self.tmpdir)

  def getExpectedXMLTree(self):
    return self.getTree(self.expectedOutput)

  @staticmethod
  def getTree(xmlFile):
    et = ElementTree()
    et.parse(xmlFile)
    return et

  @staticmethod
  def findAttributeByName(elements, processorName):
    """check if processor name is executed in tree."""
    for elem in elements:
      if processorName == elem.attrib.get('name'):
        return True
    return False

  def findProcessorInTree(self, processorName):
    """check if the testTree and the expectedTree containt the same processor with optional parameter and optional value comparison."""
    expected = TestPrepareMarlinXMLFile.findAttributeByName(
        self.expectedTree.findall('execute/processor'), processorName)
    tested = TestPrepareMarlinXMLFile.findAttributeByName(
        self.testedTree.findall('execute/processor'), processorName)

    if expected:
      print("found", processorName, "in expectedProcessors")
    if tested:
      print("found", processorName, "in testedProcessors")

    return expected == tested

  def checkProcessorParameter(self, processorName, processorParameter,
                               parameterValue):
    """check that the processorParameter has the correct value."""
    processors = self.testedTree.findall('processor')
    for proc in processors:
      if proc.attrib.get('name') == processorName:
        parameters = proc.findall('parameter')
        for param in parameters:
          if param.attrib.get('name') == processorParameter:
            if param.text is not None and param.text.strip() == str(parameterValue):
              return True
            elif param.attrib.get('value') == str(parameterValue):
              return True
            elif not parameterValue and not (param.attrib.get('value') or param.text):
              return True
            else:
              print("Parameter", processorParameter, "value is",
                    repr(param.attrib.get('value')),
                    "or", repr(param.text))

    groups = self.testedTree.findall('group')
    for group in groups:
      groupParameters = group.findall('parameter')
      groupProcessors = group.findall('processor')
      for proc in groupProcessors:
        if proc.attrib.get('name') == processorName:
          parameters = proc.findall('parameter') + groupParameters
          for param in parameters:
            if param.attrib.get('name') == processorParameter:
              if param.text is not None and param.text.strip() == str(parameterValue):
                return True
              elif param.attrib.get('value') == str(parameterValue):
                return True
              elif not parameterValue and not (param.attrib.get('value') or param.text):
                return True
              else:
                print("Parameter", processorParameter, "value is",
                      repr(param.attrib.get('value')),
                      "or", repr(param.text))

    return False

  def checkGlobalTag(self, parameterName, parameterValue):
    """check if global parameter has correct value in testedTree."""
    parameterValue = str(parameterValue)
    params = self.testedTree.findall('global/parameter')
    print("checking tag", parameterName)
    for param in params:
      if param.get('name') == parameterName:
        if param.get('value') == parameterValue or param.text == parameterValue:
          print("found parameter", parameterName, "with", parameterValue)
          return True, None

        message = "The parameter " + repr(parameterName) + " does not have the value " + repr(parameterValue) + " but " + \
          repr(param.get('value')) + " or" + repr(param.text)
        return False, message
    return False, "parameter " + parameterName + " not found "


class TestPrepareMarlinXMLFile(TestPrepareMarlinXMLFileBase):
  """tests for the perpareXMLFile function used to modify marlin steering files."""

  def setUp(self):
    self.basedir = os.getcwd()
    self.tmpdir = tempfile.mkdtemp("", dir="./")
    os.chdir(self.tmpdir)
    self.inputXMLFile = "marlininput.xml"
    self.expectedOutput = os.path.join(os.getenv('ILCDIRAC_BASE_FOLDER'),
                                       'Tests/Files/marlinExpectedOutput.xml')
    for xmlFile in ("marlininput.xml",
                    "marlinInputSmall.xml",
                    "marlinRecoProd.xml",
                    "marlininputild.xml",
                    "marlininputildNoOver.xml"):
      shutil.copyfile(os.path.join(os.getenv('ILCDIRAC_BASE_FOLDER'), 'Tests/Files/',
                                     xmlFile), xmlFile)
    # self.createInputXMLFile()
    self.expectedTree = self.getExpectedXMLTree()
    self.testedTree = None

  @patch("ILCDIRAC.Core.Utilities.MarlinXML.getOverlayFiles", new=Mock(return_value=["file1", "file2"]))
  def test_createFile(self):
    from ILCDIRAC.Core.Utilities.PrepareOptionFiles import prepareXMLFile
    res = prepareXMLFile(finalxml="outputfile.xml",
                          inputXML="marlininput.xml",
                          inputGEAR="gearMyFile.xml",
                          inputSLCIO="mySLCIOInput.slcio",
                          numberofevts=501,
                          outputFile="mySLCIOOutput.slcio",
                          outputREC="outputrec.slcio",
                          outputDST="outputdst.slcio",
                          debug=False,
                          dd4hepGeoFile="/cvmfs/monty.python.fr/myDetector.xml",
                          overlayParam=[('gghad', 0, None)],
                        )
    self.assertTrue(res['OK'])
    self.testedTree = TestPrepareMarlinXMLFile.getTree("outputfile.xml")
    self.assertTrue(self.findProcessorInTree("InitDD4hep"), "Problem with InitDD4hep")
    self.assertTrue(*self.checkGlobalTag("LCIOInputFiles", "mySLCIOInput.slcio"))
    self.assertTrue(*self.checkGlobalTag("MaxRecordNumber", 501))
    self.assertTrue(*self.checkGlobalTag("SkipNEvents", 0))
    self.assertTrue(*self.checkGlobalTag("GearXMLFile", "gearMyFile.xml"))
    self.assertTrue(*self.checkGlobalTag("Verbosity", "SILENT"))
    self.assertTrue(self.checkProcessorParameter(
        "InitDD4hep", "DD4hepXMLFile", "/cvmfs/monty.python.fr/myDetector.xml"))
    self.assertTrue(self.checkProcessorParameter("MyOverlayTiming",
                                                   "BackgroundFileNames",
                                                   "file1\nfile2"))
    # Make sure these checks are not always true
    self.assertFalse(*self.checkGlobalTag("Verbosity", "NotSILENT"))
    self.assertFalse(self.checkProcessorParameter("MyOverlayTiming",
                                                    "BackgroundFileNames",
                                                    "NotTheseFiles"))

  @patch("ILCDIRAC.Core.Utilities.MarlinXML.getOverlayFiles", new=Mock(return_value=["file1", "file2"]))
  def test_createFile_initialParametersMissing(self):
    from ILCDIRAC.Core.Utilities.PrepareOptionFiles import prepareXMLFile
    res = prepareXMLFile(finalxml="outputfilesmall.xml",
                          inputXML="marlinInputSmall.xml",
                          inputGEAR="gearMyFile.xml",
                          inputSLCIO="mySLCIOInput.slcio",
                          numberofevts=501,
                          outputFile="mySLCIOOutput.slcio",
                          outputREC="outputrec.slcio",
                          outputDST="outputdst.slcio",
                          debug=True,
                          dd4hepGeoFile="/cvmfs/monty.python.fr/myDetector.xml")
    self.assertTrue(res['OK'], res.get('Message'))
    self.testedTree = TestPrepareMarlinXMLFile.getTree("outputfilesmall.xml")
    self.assertTrue(self.findProcessorInTree("InitDD4hep"),
                     "Problem with InitDD4hep")
    self.assertTrue(*self.checkGlobalTag("LCIOInputFiles", "mySLCIOInput.slcio"))
    self.assertTrue(*self.checkGlobalTag("MaxRecordNumber", 501))
    self.assertTrue(*self.checkGlobalTag("SkipNEvents", 0))
    self.assertTrue(*self.checkGlobalTag("GearXMLFile", "gearMyFile.xml"))
    self.assertTrue(*self.checkGlobalTag("Verbosity", "WARNING"))
    self.assertTrue(self.checkProcessorParameter(
        "InitDD4hep", "DD4hepXMLFile", "/cvmfs/monty.python.fr/myDetector.xml"))
    self.assertTrue(self.checkProcessorParameter("MyOverlayTiming",
                                                   "NBunchtrain", "0"))
    self.assertTrue(self.checkProcessorParameter("MyOverlayTiming",
                                                   "NumberBackground", "0.0"))
    self.assertTrue(self.checkProcessorParameter("MyOverlayTiming",
                                                   "BackgroundFileNames", ""))
    self.assertTrue(self.checkProcessorParameter(
        "MyLCIOOutputProcessor", "LCIOOutputFile", "mySLCIOOutput.slcio"))

  @patch("ILCDIRAC.Core.Utilities.MarlinXML.getOverlayFiles", new=Mock(return_value=["file1", "file2"]))
  def test_createFile_productionOutputFiles(self):
    from ILCDIRAC.Core.Utilities.PrepareOptionFiles import prepareXMLFile
    res = prepareXMLFile(finalxml="outputprod.xml",
                          inputXML="marlinRecoProd.xml",
                          inputGEAR="gearMyFile.xml",
                          inputSLCIO="mySLCIOInput.slcio",
                          numberofevts=501,
                          outputFile='',
                          outputREC="outputrec.slcio",
                          outputDST="outputdst.slcio",
                          debug=True,
                          dd4hepGeoFile="/cvmfs/monty.python.fr/myDetector.xml",
                          overlayParam=[('gghad', 0, None)],
                        )
    self.assertTrue(res['OK'], res.get('Message'))
    self.testedTree = TestPrepareMarlinXMLFile.getTree("outputprod.xml")
    self.assertTrue(self.findProcessorInTree("InitDD4hep"),
                     "Problem with InitDD4hep")
    self.assertTrue(*self.checkGlobalTag("LCIOInputFiles", "mySLCIOInput.slcio"))
    self.assertTrue(*self.checkGlobalTag("MaxRecordNumber", 501))
    self.assertTrue(*self.checkGlobalTag("SkipNEvents", 0))
    self.assertTrue(*self.checkGlobalTag("GearXMLFile", "gearMyFile.xml"))
    self.assertTrue(*self.checkGlobalTag("Verbosity", "WARNING"))
    self.assertTrue(self.checkProcessorParameter(
        "InitDD4hep", "DD4hepXMLFile", "/cvmfs/monty.python.fr/myDetector.xml"))
    self.assertTrue(self.checkProcessorParameter(
        "MyOverlayTiming", "BackgroundFileNames", "file1\nfile2"))
    self.assertTrue(self.checkProcessorParameter(
        "MyLCIOOutputProcessor", "LCIOOutputFile", "outputrec.slcio"))
    self.assertTrue(self.checkProcessorParameter(
        "DSTOutput", "LCIOOutputFile", "outputdst.slcio"))

  @patch("ILCDIRAC.Core.Utilities.MarlinXML.getOverlayFiles", new=Mock(return_value=["file1", "file2"]))
  def test_createFile_ildRecoFile(self):
    from ILCDIRAC.Core.Utilities.PrepareOptionFiles import prepareXMLFile
    res = prepareXMLFile(finalxml="outputprodild.xml",
                          inputXML="marlininputild.xml",
                          inputGEAR="gearMyFile.xml",
                          inputSLCIO="mySLCIOInput.slcio",
                          numberofevts=501,
                          outputFile='',
                          outputREC="outputrec.slcio",
                          outputDST="outputdst.slcio",
                          debug=True,
                          dd4hepGeoFile="/cvmfs/monty.python.fr/myDetector.xml",
                          overlayParam=[('aa_lowpt', 333, None)],
                        )
    self.assertTrue(res['OK'], res.get('Message'))
    self.testedTree = TestPrepareMarlinXMLFile.getTree("outputprodild.xml")
    #self.assertTrue( self.findProcessorInTree( "BGoverLay" ), "Problem with InitDD4hep" )
    self.assertTrue(*self.checkGlobalTag("LCIOInputFiles", "mySLCIOInput.slcio"))
    self.assertTrue(*self.checkGlobalTag("MaxRecordNumber", 501))
    self.assertTrue(*self.checkGlobalTag("SkipNEvents", 0))
    self.assertTrue(*self.checkGlobalTag("GearXMLFile", "gearMyFile.xml"))
    self.assertTrue(*self.checkGlobalTag("Verbosity", "WARNING"))
    self.assertTrue(self.checkProcessorParameter("BgOverlay", "InputFileNames",
                                                   "file1\nfile2"))
    self.assertTrue(self.checkProcessorParameter("BgOverlay", "NSkipEventsRandom",
                                                   "666"))

  @patch("ILCDIRAC.Core.Utilities.MarlinXML.getOverlayFiles", new=Mock(return_value=["file1", "file2"]))
  def test_createFile_ildRecoFile_NoOver(self):
    from ILCDIRAC.Core.Utilities.PrepareOptionFiles import prepareXMLFile
    res = prepareXMLFile(finalxml="outputprodildNoOver.xml",
                          inputXML="marlininputildNoOver.xml",
                          inputGEAR="gearMyFile.xml",
                          inputSLCIO="mySLCIOInput.slcio",
                          numberofevts=501,
                          outputFile='',
                          outputREC="outputrec.slcio",
                          outputDST="outputdst.slcio",
                          debug=True,
                          dd4hepGeoFile="/cvmfs/monty.python.fr/myDetector.xml")
    self.assertTrue(res['OK'], res.get('Message'))
    self.testedTree = TestPrepareMarlinXMLFile.getTree("outputprodildNoOver.xml")
    #self.assertTrue( self.findProcessorInTree( "BGoverLay" ), "Problem with InitDD4hep" )
    self.assertTrue(*self.checkGlobalTag("LCIOInputFiles", "mySLCIOInput.slcio"))
    self.assertTrue(*self.checkGlobalTag("MaxRecordNumber", 501))
    self.assertTrue(*self.checkGlobalTag("SkipNEvents", 0))
    self.assertTrue(*self.checkGlobalTag("GearXMLFile", "gearMyFile.xml"))
    self.assertTrue(*self.checkGlobalTag("Verbosity", "WARNING"))
    self.assertTrue(self.checkProcessorParameter("BgOverlay", "InputFileNames", ""))
    self.assertTrue(self.checkProcessorParameter("BgOverlay", "expBG", "0.0"))

  def test_createFile_NoOutputSet(self):
    from ILCDIRAC.Core.Utilities.PrepareOptionFiles import prepareXMLFile
    res = prepareXMLFile(finalxml="marlinNoOutput.xml",
                          inputXML="marlinInputSmall.xml",
                          inputGEAR="gearMyFile.xml",
                          inputSLCIO="mySLCIOInput.slcio",
                          numberofevts=501,
                          outputFile='',
                          outputREC='',
                          outputDST='',
                          debug=True,
                          dd4hepGeoFile="/cvmfs/monty.python.fr/myDetector.xml")
    self.assertTrue(res['OK'], res.get('Message'))
    self.testedTree = TestPrepareMarlinXMLFile.getTree("marlinNoOutput.xml")
    #self.assertTrue( self.findProcessorInTree( "BGoverLay" ), "Problem with InitDD4hep" )
    self.assertTrue(*self.checkGlobalTag("LCIOInputFiles", "mySLCIOInput.slcio"))
    self.assertTrue(*self.checkGlobalTag("MaxRecordNumber", 501))
    self.assertTrue(*self.checkGlobalTag("SkipNEvents", 0))
    self.assertTrue(*self.checkGlobalTag("GearXMLFile", "gearMyFile.xml"))
    self.assertTrue(*self.checkGlobalTag("Verbosity", "WARNING"))
    self.assertTrue(self.checkProcessorParameter("MyLCIOOutputProcessor",
                                                   "LCIOOutputFile", ""))

  @patch("ILCDIRAC.Core.Utilities.MarlinXML.getOverlayFiles", new=Mock(return_value=[]))
  def test_createFile_overlayTiming_NoFiles(self):
    from ILCDIRAC.Core.Utilities.PrepareOptionFiles import prepareXMLFile
    res = prepareXMLFile(finalxml="outputprod.xml",
                          inputXML="marlinRecoProd.xml",
                          inputGEAR="gearMyFile.xml",
                          inputSLCIO="mySLCIOInput.slcio",
                          numberofevts=501,
                          outputFile='',
                          outputREC="outputrec.slcio",
                          outputDST="outputdst.slcio",
                          debug=True,
                          dd4hepGeoFile="/cvmfs/monty.python.fr/myDetector.xml",
                          overlayParam=[('gghad', 0, None)],
                        )
    self.assertFalse(res['OK'])
    self.assertIn("Could not find any overlay files", res['Message'])

  @patch("ILCDIRAC.Core.Utilities.MarlinXML.getOverlayFiles", new=Mock(return_value=[]))
  def test_createFile_bgoverlay_NoFiles(self):
    from ILCDIRAC.Core.Utilities.PrepareOptionFiles import prepareXMLFile
    res = prepareXMLFile(finalxml="outputprod.xml",
                          inputXML="marlininputild.xml",
                          inputGEAR="gearMyFile.xml",
                          inputSLCIO="mySLCIOInput.slcio",
                          numberofevts=501,
                          outputFile='',
                          outputREC="outputrec.slcio",
                          outputDST="outputdst.slcio",
                          debug=True,
                          dd4hepGeoFile="/cvmfs/monty.python.fr/myDetector.xml",
                          overlayParam=[('aa_lowpt', 0, None)],
                        )
    self.assertFalse(res['OK'])
    self.assertIn("Could not find any overlay files", res['Message'])

  @patch("ILCDIRAC.Core.Utilities.MarlinXML.getOverlayFiles", new=Mock(return_value=[]))
  def test_createFile_bgoverlay_failOverlayActive(self):
    from ILCDIRAC.Core.Utilities.PrepareOptionFiles import prepareXMLFile
    res = prepareXMLFile(finalxml="outputprod.xml",
                          inputXML="marlininputild.xml",
                          inputGEAR="gearMyFile.xml",
                          inputSLCIO="mySLCIOInput.slcio",
                          numberofevts=501,
                          outputFile='',
                          outputREC="outputrec.slcio",
                          outputDST="outputdst.slcio",
                          debug=True,
                          dd4hepGeoFile="/cvmfs/monty.python.fr/myDetector.xml",
                          overlayParam=[],
                        )
    assertDiracFailsWith(res, 'Found active overlay processors', self)

  def test_createFile_parseError(self):

    def parseModified(*_args, **_kwargs):
      """throw exception for parse."""
      raise RuntimeError("Parse Failed")

    from ILCDIRAC.Core.Utilities.PrepareOptionFiles import prepareXMLFile
    with patch("%s.ElementTree.parse" % MODULE_NAME, new=parseModified):
      res = prepareXMLFile(finalxml="outputprod.xml",
                            inputXML="marlininputild.xml",
                            inputGEAR="gearMyFile.xml",
                            inputSLCIO="mySLCIOInput.slcio",
                            numberofevts=501,
                            outputFile='',
                            outputREC="outputrec.slcio",
                            outputDST="outputdst.slcio",
                            debug=True,
                            dd4hepGeoFile="/cvmfs/monty.python.fr/myDetector.xml")
    self.assertFalse(res['OK'])
    self.assertIn("Found Exception when parsing", res['Message'])

  def test_createFile_inputFaulty(self):
    from ILCDIRAC.Core.Utilities.PrepareOptionFiles import prepareXMLFile
    res = prepareXMLFile(finalxml="outputprod.xml",
                          inputXML="marlininputild.xml",
                          inputGEAR="gearMyFile.xml",
                          inputSLCIO=None,
                          numberofevts=501,
                          outputFile='',
                          outputREC="outputrec.slcio",
                          outputDST="outputdst.slcio",
                          debug=True,
                          dd4hepGeoFile="/cvmfs/monty.python.fr/myDetector.xml")
    self.assertFalse(res['OK'], res['Message'])
    self.assertIn("inputSLCIO is neither", res['Message'])

  @patch("ILCDIRAC.Core.Utilities.MarlinXML.getOverlayFiles", new=Mock(return_value=["file1", "file2"]))
  def test_createFile_inputList(self):
    from ILCDIRAC.Core.Utilities.PrepareOptionFiles import prepareXMLFile
    res = prepareXMLFile(finalxml="outputfile.xml",
                          inputXML="marlininput.xml",
                          inputGEAR="gearMyFile.xml",
                          inputSLCIO=["file1", "file2"],
                          numberofevts=501,
                          outputFile="mySLCIOOutput.slcio",
                          outputREC="outputrec.slcio",
                          outputDST="outputdst.slcio",
                          debug=False,
                          dd4hepGeoFile="/cvmfs/monty.python.fr/myDetector.xml",
                          overlayParam=[('gghad', 0, None)])
    self.assertTrue(res['OK'], res.get('Message', ''))
    self.testedTree = TestPrepareMarlinXMLFile.getTree("outputfile.xml")
    self.assertTrue(self.findProcessorInTree("InitDD4hep"),
                     "Problem with InitDD4hep")
    self.assertTrue(*self.checkGlobalTag("LCIOInputFiles", "file1 file2"))
    self.assertTrue(*self.checkGlobalTag("MaxRecordNumber", 501))
    self.assertTrue(*self.checkGlobalTag("SkipNEvents", 0))
    self.assertTrue(*self.checkGlobalTag("GearXMLFile", "gearMyFile.xml"))
    self.assertTrue(*self.checkGlobalTag("Verbosity", "SILENT"))


class TestPrepareClicProdXMLFile(TestPrepareMarlinXMLFileBase):
  """tests for the perpareXMLFile function used to modify marlin steering files."""

  def setUp(self):
    self.basedir = os.getcwd()
    self.tmpdir = tempfile.mkdtemp("", dir="./")
    os.chdir(self.tmpdir)
    for xmlFile in ("clicReconstruction.xml",
                   ):
      shutil.copyfile(os.path.join(os.getenv('ILCDIRAC_BASE_FOLDER'), 'Tests', 'Files',
                                     xmlFile), xmlFile)
    self.testedTree = None

  @patch("ILCDIRAC.Core.Utilities.MarlinXML.getOverlayFiles", new=Mock(return_value=["file1", "file2"]))
  def test_createFile_clicProd2017(self):
    from ILCDIRAC.Core.Utilities.PrepareOptionFiles import prepareXMLFile
    res = prepareXMLFile(finalxml="outputprod.xml",
                          inputXML="clicReconstruction.xml",
                          inputGEAR="",
                          inputSLCIO="mySLCIOInput.slcio",
                          numberofevts=501,
                          outputFile='',
                          outputREC="outputrec.slcio",
                          outputDST="outputdst.slcio",
                          debug=False,
                          dd4hepGeoFile="/cvmfs/monty.python.fr/myDetector.xml",
                          overlayParam=[('gghad', 0, 'Overlay380GeV')],
                        )
    self.assertTrue(res['OK'], res.get('Message'))
    self.testedTree = TestPrepareMarlinXMLFile.getTree("outputprod.xml")
    self.assertTrue(*self.checkGlobalTag("LCIOInputFiles", "mySLCIOInput.slcio"))
    self.assertTrue(*self.checkGlobalTag("MaxRecordNumber", 501))
    self.assertTrue(*self.checkGlobalTag("SkipNEvents", 0))
    self.assertTrue(*self.checkGlobalTag("Verbosity", "SILENT"))
    self.assertTrue(self.checkProcessorParameter(
        "InitDD4hep", "DD4hepXMLFile", "/cvmfs/monty.python.fr/myDetector.xml"))
    self.assertTrue(self.checkProcessorParameter(
        "Overlay380GeV", "BackgroundFileNames", "file1\nfile2"))
    self.assertTrue(self.checkProcessorParameter(
        "Output_REC", "LCIOOutputFile", "outputrec.slcio"))
    self.assertTrue(self.checkProcessorParameter(
        "Output_DST", "LCIOOutputFile", "outputdst.slcio"))

  @patch("ILCDIRAC.Core.Utilities.MarlinXML.getOverlayFiles", new=Mock(return_value=[]))
  def test_createFile_clicProd2017_noOverlayFiles(self):
    from ILCDIRAC.Core.Utilities.PrepareOptionFiles import prepareXMLFile
    res = prepareXMLFile(finalxml="outputprod.xml",
                          inputXML="clicReconstruction.xml",
                          inputGEAR="",
                          inputSLCIO="mySLCIOInput.slcio",
                          numberofevts=501,
                          outputFile='',
                          outputREC="outputrec.slcio",
                          outputDST="outputdst.slcio",
                          debug=False,
                          dd4hepGeoFile="/cvmfs/monty.python.fr/myDetector.xml",
                          overlayParam=[('gghad', 0, 'Overlay380GeV')],
                        )
    assertDiracFailsWith(res, 'Could not find any overlay Files', self)


if __name__ == "__main__":
  SUITE = unittest.defaultTestLoader.loadTestsFromTestCase(TestPrepareClicProdXMLFile)
  TESTRESULT = unittest.TextTestRunner(verbosity=2).run(SUITE)
  print(TESTRESULT)



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Tests for the ResolveSE module."""
from __future__ import absolute_import
import unittest
from mock import patch, MagicMock as Mock

from DIRAC import S_OK, S_ERROR
from ILCDIRAC.Core.Utilities.ResolveSE import getDestinationSEList
from Tests.Utilities.GeneralUtils import assertDiracFailsWith, assertDiracSucceedsWith, assertDiracSucceedsWith_equals

__RCSID__ = "$Id$"

MODULE_NAME = 'ILCDIRAC.Core.Utilities.ResolveSE'


class TestGetDestinationSE(unittest.TestCase):
  """Test getDestinationSEList function."""

  def test_getdestinationse(self):
    options_dict = {'/Resources/StorageElements/myOutputSE': S_ERROR('bla')}
    values_dict = {'/Resources/Sites/pre/pre.myTestSite.country/AssociatedSEs/myOutputSE': [],
                    '/Resources/StorageElementGroups/myOutputSE': ['some_grp_member', 'GroupElement', 'StorEl', 'LocalElement']}
    conf_mock = Mock()
    conf_mock.getOptions.side_effect = lambda path: options_dict[path]
    conf_mock.getValue.side_effect = lambda path, _: values_dict[path]
    with patch('%s.getSEsForSite' % MODULE_NAME, new=Mock(return_value=S_OK(['StorEl', 'LocalElement']))), \
         patch('%s.gConfig' % MODULE_NAME, conf_mock):
      result = getDestinationSEList('myOutputSE', 'pre.myTestSite.country', 'AnyOutputmodeTest')
      assertDiracSucceedsWith_equals(result, ['StorEl', 'LocalElement', 'some_grp_member', 'GroupElement'], self)

  def test_getdestinationse_alias_found(self):
    options_dict = {'/Resources/StorageElements/myOutputSE': S_ERROR('bla')}
    values_dict = {'/Resources/Sites/pre/pre.myTestSite.country/AssociatedSEs/myOutputSE':
                    'myTest_Alias_value'}
    conf_mock = Mock()
    conf_mock.getOptions.side_effect = lambda path: options_dict[path]
    conf_mock.getValue.side_effect = lambda path, _: values_dict[path]
    with patch('%s.gConfig' % MODULE_NAME, conf_mock):
      result = getDestinationSEList('myOutputSE', 'pre.myTestSite.country', 'AnyOutputmodeTest')
      assertDiracSucceedsWith(result, 'myTest_Alias_value', self)

  def test_getdestinationse_nogroupse(self):
    options_dict = {'/Resources/StorageElements/myOutputSE': S_ERROR('bla')}
    values_dict = {'/Resources/Sites/pre/pre.myTestSite.country/AssociatedSEs/myOutputSE': [],
                    '/Resources/StorageElementGroups/myOutputSE': []}
    conf_mock = Mock()
    conf_mock.getOptions.side_effect = lambda path: options_dict[path]
    conf_mock.getValue.side_effect = lambda path, _: values_dict[path]
    #DIRAC.gConfig = conf_mock
    with patch('%s.getSEsForSite' % MODULE_NAME, new=Mock(return_value=S_OK(['myTestStorageElement1', 'otherStorageElem2']))), \
         patch('%s.gConfig' % MODULE_NAME, conf_mock):
      result = getDestinationSEList('myOutputSE', 'pre.myTestSite.country', 'AnyOutputmodeTest')
      assertDiracFailsWith(result, 'failed to resolve se myoutputse', self)

  def test_getdestinationse_localgroupse(self):
    options_dict = {'/Resources/StorageElements/myOutputSE': S_ERROR('bla')}
    values_dict = {'/Resources/Sites/pre/pre.myTestSite.country/AssociatedSEs/myOutputSE': [],
                    '/Resources/StorageElementGroups/myOutputSE': ['some_elements', 'ignore_me',
                                                                     'otherStorageElem2']}
    conf_mock = Mock()
    conf_mock.getOptions.side_effect = lambda path: options_dict[path]
    conf_mock.getValue.side_effect = lambda path, _: values_dict[path]
    with patch('%s.getSEsForSite' % MODULE_NAME, new=Mock(return_value=S_OK(['myTestStorageElement1', 'otherStorageElem2']))), \
         patch('%s.gConfig' % MODULE_NAME, conf_mock):
      result = getDestinationSEList('myOutputSE', 'pre.myTestSite.country', 'Local')
      assertDiracSucceedsWith_equals(result, ['otherStorageElem2'], self)

  def test_getdestinationse_associatedse(self):
    options_dict = {'/Resources/StorageElements/myOutputSE': S_ERROR('bla')}
    values_dict = {'/Resources/Sites/pre/pre.myTestSite.country/AssociatedSEs/myOutputSE': [],
                    '/Resources/StorageElementGroups/myOutputSE': ['some_elements', 'ignore_me'],
                    '/Resources/Countries/country/AssociatedSEs/myOutputSE': 'myassociated_se'}
    conf_mock = Mock()
    conf_mock.getOptions.side_effect = lambda path: options_dict[path]
    conf_mock.getValue.side_effect = lambda path, _: values_dict[path]
    with patch('%s.getSEsForSite' % MODULE_NAME, new=Mock(return_value=S_OK(['myTestStorageElement1', 'otherStorageElem2']))), \
         patch('%s.gConfig' % MODULE_NAME, conf_mock):
      result = getDestinationSEList('myOutputSE', 'pre.myTestSite.country', 'Local')
      assertDiracSucceedsWith_equals(result, ['myassociated_se'], self)

  def test_getdestinationse_noassoc_country(self):
    options_dict = {'/Resources/StorageElements/myOutputSE': S_ERROR('bla')}
    option_dict = {'/Resources/Countries/country/AssignedTo': S_OK('MyTestCountry.Assigned'),
                    '/Resources/Countries/MyTestCountry.Assigned/AssociatedSEs': S_OK('something')}
    values_dict = {'/Resources/Sites/pre/pre.myTestSite.country/AssociatedSEs/myOutputSE': [],
                    '/Resources/StorageElementGroups/myOutputSE': ['some_elements', 'ignore_me'],
                    '/Resources/Countries/MyTestCountry.Assigned/AssociatedSEs/myOutputSE':
                    ['myassociated_alias_se'],
                    '/Resources/Countries/country/AssociatedSEs/myOutputSE': ''}
    conf_mock = Mock()
    conf_mock.getOptions.side_effect = lambda path: options_dict[path]
    conf_mock.getOption.side_effect = lambda path: option_dict[path]
    conf_mock.getValue.side_effect = lambda path, _: values_dict[path]
    with patch('%s.getSEsForSite' % MODULE_NAME, new=Mock(return_value=S_OK(['myTestStorageElement1', 'otherStorageElem2']))), \
         patch('%s.gConfig' % MODULE_NAME, conf_mock):
      result = getDestinationSEList('myOutputSE', 'pre.myTestSite.country', 'Local')
      assertDiracSucceedsWith_equals(result, ['myassociated_alias_se'], self)

  def test_getdestinationse_no_assoc_list(self):
    options_dict = {'/Resources/StorageElements/myOutputSE': S_ERROR('bla')}
    option_dict = {'/Resources/Countries//AssignedTo': S_OK('')}
    values_dict = {'/Resources/Sites/pre/pre.myTestSite./AssociatedSEs/myOutputSE': [],
                    '/Resources/StorageElementGroups/myOutputSE': ['some_elements', 'ignore_me'],
                    '/Resources/Countries/MyTestCountry.Assigned/AssociatedSEs/myOutputSE':
                    ['myassociated_alias_se'],
                    '/Resources/Countries//AssociatedSEs/myOutputSE': ''}
    conf_mock = Mock()
    conf_mock.getOptions.side_effect = lambda path: options_dict[path]
    conf_mock.getOption.side_effect = lambda path: option_dict[path]
    conf_mock.getValue.side_effect = lambda path, _: values_dict[path]
    with patch('%s.getSEsForSite' % MODULE_NAME, new=Mock(return_value=S_OK(['myTestStorageElement1', 'otherStorageElem2']))), \
         patch('%s.gConfig' % MODULE_NAME, conf_mock):
      result = getDestinationSEList('myOutputSE', 'pre.myTestSite.', 'Local')
      assertDiracFailsWith(result, 'could not determine associated se list', self)

  def test_getdestinationse_no_assoc_alias(self):
    options_dict = {'/Resources/StorageElements/myOutputSE': S_ERROR('bla')}
    option_dict = {'/Resources/Countries/country/AssignedTo': S_OK('MyTestCountry.Assigned'),
                    '/Resources/Countries/MyTestCountry.Assigned/AssociatedSEs': S_OK('something')}
    values_dict = {'/Resources/Sites/pre/pre.myTestSite.country/AssociatedSEs/myOutputSE': [],
                    '/Resources/StorageElementGroups/myOutputSE': ['some_elements', 'ignore_me'],
                    '/Resources/Countries/MyTestCountry.Assigned/AssociatedSEs/myOutputSE': [],
                    '/Resources/Countries/country/AssociatedSEs/myOutputSE': ''}
    conf_mock = Mock()
    conf_mock.getOptions.side_effect = lambda path: options_dict[path]
    conf_mock.getOption.side_effect = lambda path: option_dict[path]
    conf_mock.getValue.side_effect = lambda path, _: values_dict[path]
    with patch('%s.getSEsForSite' % MODULE_NAME, new=Mock(return_value=S_OK(['myTestStorageElement1', 'otherStorageElem2']))), \
         patch('%s.gConfig' % MODULE_NAME, conf_mock):
      result = getDestinationSEList('myOutputSE', 'pre.myTestSite.country', 'Local')
      assertDiracFailsWith(result, 'failed to resolve se', self)



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Tests for OverlayDB."""
from __future__ import absolute_import
import unittest
from mock import patch, MagicMock as Mock

from Tests.Utilities.GeneralUtils import assertDiracFailsWith, assertDiracSucceeds, \
    assertDiracSucceedsWith_equals, assertMockCalls
from DIRAC import S_OK, S_ERROR

__RCSID__ = "$Id$"

MODULE_NAME = 'ILCDIRAC.OverlaySystem.DB.OverlayDB'

# pylint: disable=no-member


class TestOverlayDB(unittest.TestCase):
  """Tests of OverlayDB."""

  def setUp(self):
    """Prepare OverlayDB object."""
    from ILCDIRAC.OverlaySystem.DB.OverlayDB import OverlayDB
    from DIRAC.Core.Base.DB import DB
    value_dict = {'/Overlay/MaxConcurrentRunning': 10, '/Overlay/Sites/testSite1/MaxConcurrentRunning': 2,
                   '/Overlay/Sites/myOtherSite/MaxConcurrentRunning': 2}
    sections_dict = {'/Overlay/Sites/': ['testSite1', 'myOtherSite']}
    self.ops_mock = Mock()
    self.ops_mock.getValue.side_effect = lambda x, _: value_dict[x]
    self.ops_mock.getSections.side_effect = lambda x: S_OK(sections_dict[x])
    with patch('%s.Operations' % MODULE_NAME, new=Mock(return_value=self.ops_mock)), \
         patch.object(OverlayDB, '_createTables', new=Mock()), \
         patch.object(DB, '__init__', new=Mock()):
      self.odb = OverlayDB()

  def test_getsites(self):
    con_mock = Mock()
    with patch('%s.OverlayDB._getConnection' % MODULE_NAME, new=Mock(return_value=S_OK(con_mock))), \
         patch('%s.OverlayDB._query' % MODULE_NAME, new=Mock(return_value=S_OK([('some_row',), ('other_row', )]))) as query_mock:
      assertDiracSucceedsWith_equals(self.odb.getSites(), ['some_row', 'other_row'], self)
      query_mock.assert_called_once_with('SELECT Site From OverlayData;', con_mock)

  def test_getsites_query_fails(self):
    con_mock = Mock()
    with patch('%s.OverlayDB._getConnection' % MODULE_NAME, new=Mock(return_value=S_OK(con_mock))), \
         patch('%s.OverlayDB._query' % MODULE_NAME, new=Mock(return_value=S_ERROR('test_query_fails'))):
      assertDiracFailsWith(self.odb.getSites(), 'Could not get sites', self)

  def test_getsites_nosites_available(self):
    con_mock = Mock()
    with patch('%s.OverlayDB._query' % MODULE_NAME, new=Mock(return_value=S_OK([]))):
      assertDiracSucceedsWith_equals(self.odb.getSites(con_mock), [], self)

  def test_getsites_connection_fails(self):
    with patch('%s.OverlayDB._getConnection' % MODULE_NAME, new=Mock(return_value=S_ERROR())), \
        patch('%s.OverlayDB._query' % MODULE_NAME, new=Mock(return_value=S_ERROR('noconnection'))):
      assertDiracFailsWith(self.odb.getSites(), 'Could not get sites', self)

  def test_setjobsatsites(self):
    con_mock = Mock()
    with patch('%s.OverlayDB._update' % MODULE_NAME, new=Mock(return_value=S_OK())) as update_mock:
      assertDiracSucceeds(self.odb.setJobsAtSites(
          {'MyTestSite1': 1487, 'other_site': '138', 'large_testsite': 40913.2}, con_mock), self)
      assertMockCalls(update_mock, [
          ("UPDATE OverlayData SET NumberOfJobs=1487 WHERE Site='MyTestSite1';", con_mock),
          ("UPDATE OverlayData SET NumberOfJobs=138 WHERE Site='other_site';", con_mock),
          ("UPDATE OverlayData SET NumberOfJobs=40913 WHERE Site='large_testsite';", con_mock)], self)

  def test_setjobsatsites_nothingtodo(self):
    assertDiracSucceeds(self.odb.setJobsAtSites({}, Mock()), self)

  def test_setjobsatsites_update_fails(self):
    def replace_update(request):
      """Mocks the return value of the _update method."""
      if 'other_site' in request:
        return S_ERROR()
      return S_OK()
    with patch('%s.OverlayDB._update' % MODULE_NAME, new=Mock(side_effect=lambda request, _: replace_update(request))):
      assertDiracFailsWith(self.odb.setJobsAtSites({'MyTestSite1': 1487,
                                                    'other_site': '138',
                                                    'large_testsite': 40913.2},
                                                   Mock()),
                           'could not set number of jobs at site',
                           self)

  def test_getjobsatsite(self):
    con_mock = Mock()
    with patch('%s.OverlayDB._query' % MODULE_NAME, new=Mock(return_value=S_OK([[123]]))) as query_mock:
      assertDiracSucceedsWith_equals(self.odb.getJobsAtSite('myTest_Site1', con_mock), 123, self)
      query_mock.assert_called_once_with("SELECT NumberOfJobs FROM OverlayData WHERE Site='myTest_Site1';", con_mock)

  def test_getjobsatsite_query_fails(self):
    con_mock = Mock()
    with patch('%s.OverlayDB._query' % MODULE_NAME, new=Mock(return_value=S_ERROR())):
      assertDiracSucceedsWith_equals(self.odb.getJobsAtSite('myTest_Site1', con_mock), 0, self)

  def test_getjobsatsite_no_site_found(self):
    con_mock = Mock()
    with patch('%s.OverlayDB._query' % MODULE_NAME, new=Mock(return_value=S_OK([]))):
      assertDiracSucceedsWith_equals(self.odb.getJobsAtSite('nonexistent_site_testme', con_mock), 0, self)

  def test_canrun_toomanyjobs(self):
    con_mock = Mock()
    with patch('%s.OverlayDB._query' % MODULE_NAME, new=Mock(return_value=S_OK([[2]]))) as query_mock:
      assertDiracSucceedsWith_equals(self.odb.canRun('testSite1', con_mock), False, self)
      query_mock.assert_called_once_with("SELECT NumberOfJobs FROM OverlayData WHERE Site='testSite1';",
                                          con_mock)

  def test_canrun_add_to_new_site(self):
    con_mock = Mock()
    with patch('%s.OverlayDB._query' % MODULE_NAME, new=Mock(return_value=S_ERROR())), \
         patch('%s.OverlayDB._update' % MODULE_NAME, new=Mock(return_value=S_OK())) as update_mock:
      assertDiracSucceedsWith_equals(self.odb.canRun('tenJobSite', con_mock), True, self)
      assertMockCalls(update_mock,
                       [("INSERT INTO OverlayData (Site,NumberOfJobs) VALUES ('tenJobSite',1);", con_mock),
                         ("UPDATE OverlayData SET NumberOfJobs=2 WHERE Site='tenJobSite';", con_mock)], self)

  def test_canrun_addsite_fails(self):
    con_mock = Mock()
    with patch('%s.OverlayDB._query' % MODULE_NAME, new=Mock(return_value=S_ERROR())), \
         patch('%s.OverlayDB._update' % MODULE_NAME, new=Mock(side_effect=[S_ERROR('update_test_err'), S_OK()])) as update_mock:
      assertDiracSucceedsWith_equals(self.odb.canRun('tenJobSite', con_mock), True, self)
      assertMockCalls(update_mock,
                       [("INSERT INTO OverlayData (Site,NumberOfJobs) VALUES ('tenJobSite',1);", con_mock),
                         ("UPDATE OverlayData SET NumberOfJobs=2 WHERE Site='tenJobSite';", con_mock)], self)

  def test_jobdone(self):
    con_mock = Mock()
    with patch('%s.OverlayDB._query' % MODULE_NAME, new=Mock(return_value=S_OK([[148]]))) as query_mock, \
         patch('%s.OverlayDB._update' % MODULE_NAME, new=Mock(return_value=S_OK())) as update_mock:
      assertDiracSucceeds(self.odb.jobDone('my_TestSite1', con_mock), self)
      query_mock.assert_called_once_with("SELECT NumberOfJobs FROM OverlayData WHERE Site='my_TestSite1';",
                                          con_mock)
      update_mock.assert_called_once_with("UPDATE OverlayData SET NumberOfJobs=147 WHERE Site='my_TestSite1';",
                                           con_mock)

  def test_jobdone_checksite_fails(self):
    con_mock = Mock()
    with patch('%s.OverlayDB._query' % MODULE_NAME, new=Mock(return_value=S_ERROR())):
      assertDiracFailsWith(self.odb.jobDone('my_TestSite1', con_mock), 'Could not get site', self)

  def test_jobdone_nojobs(self):
    con_mock = Mock()
    with patch('%s.OverlayDB._query' % MODULE_NAME, new=Mock(return_value=S_OK([[1]]))) as query_mock:
      assertDiracSucceeds(self.odb.jobDone('my_TestSite1', con_mock), self)
      query_mock.assert_called_once_with("SELECT NumberOfJobs FROM OverlayData WHERE Site='my_TestSite1';",
                                          con_mock)

  def test_jobdone_update_fails(self):
    con_mock = Mock()
    with patch('%s.OverlayDB._query' % MODULE_NAME, new=Mock(return_value=S_OK([[148]]))), \
         patch('%s.OverlayDB._update' % MODULE_NAME, new=Mock(return_value=S_ERROR('update_test_err'))):
      assertDiracFailsWith(self.odb.jobDone('my_TestSite1', con_mock), 'update_test_err', self)



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""User Job class. Used to define user jobs!

Example usage:

>>> from ILCDIRAC.Interfaces.API.NewInterface.UserJob import UserJob
>>> from ILCDIRAC.Interfaces.API.DiracILC import DiracILC
>>> myDiracInstance = DiracILC( withRepo=False )
>>> myJob = UserJob()
>>> ...
>>> myJob.append( myMarlinApp )
>>> myJob.submit(myDiracInstance)

:author: Stephane Poss
:author: Remi Ete
:author: Ching Bon Lam
"""

from __future__ import absolute_import
from DIRAC import S_OK, gLogger
from DIRAC.Core.Security.ProxyInfo import getProxyInfo

from ILCDIRAC.Interfaces.API.NewInterface.Job import Job
from ILCDIRAC.Interfaces.API.DiracILC import DiracILC
from ILCDIRAC.Interfaces.Utilities.SplitMixin import SplitMixin
import six


__RCSID__ = "$Id$"
LOG = gLogger.getSubLogger(__name__)


class UserJob(SplitMixin, Job):
  """User job class.

  To be used by users, not for production.
  """

  def __init__(self, script=None):
    """Initialize UserJob, including proxy and splitmixin."""
    super(UserJob, self).__init__(script)
    self.type = 'User'
    self.diracinstance = None
    self.usergroup = ['ilc_user', 'calice_user', 'fcc_user']
    self.proxyinfo = getProxyInfo()
    SplitMixin._initialize(self)

  def submit(self, diracinstance=None, mode="wms"):
    """Submit call: when your job is defined, and all applications are set, you need to call this to add the job to DIRAC.

    :param diracinstance: DiracILC instance
    :type diracinstance: ~ILCDIRAC.Interfaces.API.DiracILC.DiracILC
    :param str mode: "wms" (default), "agent", or "local"

    .. note ::
      The *local* mode means that the job will be run on the submission machine. Use this mode for testing of
      submission scripts
    """
    if self._splittingOption:
      result = self._split()
      if 'OK' in result and not result['OK']:
        return result

    # Check the credentials. If no proxy or not user proxy, return an error
    if not self.proxyinfo['OK']:
      LOG.error("Not allowed to submit a job, you need a %s proxy." % self.usergroup)
      return self._reportError("Not allowed to submit a job, you need a %s proxy." % self.usergroup,
                               self.__class__.__name__)
    if 'group' in self.proxyinfo['Value']:
      group = self.proxyinfo['Value']['group']
      if group not in self.usergroup:
        LOG.error("Not allowed to submit a job, you need a %s proxy." % self.usergroup)
        return self._reportError("Not allowed to submit job, you need a %s proxy." % self.usergroup,
                                 self.__class__.__name__)
    else:
      LOG.error("Could not determine group, you do not have the right proxy.")
      return self._reportError("Could not determine group, you do not have the right proxy.")

    res = self._addToWorkflow()
    if not res['OK']:
      return res
    self.oktosubmit = True
    if not diracinstance:
      self.diracinstance = DiracILC()
    else:
      self.diracinstance = diracinstance
    return self.diracinstance.submitJob(self, mode)

  #############################################################################
  def setInputData(self, lfns):
    """Specify input data by Logical File Name (LFN).

    Input files specified via this function will be automatically staged if necessary.

    Example usage:

    >>> job = UserJob()
    >>> job.setInputData(['/ilc/prod/whizard/processlist.whiz'])

    :param lfns: Logical File Names
    :type lfns: Single LFN string or list of LFNs
    """
    if isinstance(lfns, list) and lfns:
      for i, lfn in enumerate(lfns):
        lfns[i] = lfn.replace('LFN:', '')
      #inputData = map( lambda x: 'LFN:' + x, lfns )
      inputData = lfns  # because we don't need the LFN: for inputData, and it breaks the
      # resolution of the metadata in the InputFilesUtilities
      inputDataStr = ';'.join(inputData)
      description = 'List of input data specified by LFNs'
      self._addParameter(self.workflow, 'InputData', 'JDL', inputDataStr, description)
    elif isinstance(lfns, six.string_types):  # single LFN
      description = 'Input data specified by LFN'
      self._addParameter(self.workflow, 'InputData', 'JDL', lfns, description)
    else:
      kwargs = {'lfns': lfns}
      return self._reportError('Expected lfn string or list of lfns for input data', **kwargs)

    return S_OK()

  def setInputSandbox(self, flist):
    """Add files to the input sandbox, can be on the local machine or on the grid.

    >>> job = UserJob()
    >>> job.setInputSandbox( ['LFN:/ilc/user/u/username/libraries.tar.gz',
    >>>                       'mySteeringFile.xml'] )

    :param flist: Files for the inputsandbox
    :type flist: `python:list` or `str`
    """
    if isinstance(flist, six.string_types):
      flist = [flist]
    if not isinstance(flist, list):
      return self._reportError("File passed must be either single file or list of files.")
    self.inputsandbox.extend(flist)
    return S_OK()

  #############################################################################
  def setOutputData(self, lfns, OutputPath='', OutputSE=''):
    """For specifying output data to be registered in Grid storage.  If a list of OutputSEs are specified the job wrapper will try each in turn until successful.

    Example usage:

    >>> job = UserJob()
    >>> job.setOutputData(['Ntuple.root'])

    :param lfns: Output data file or files
    :type lfns: Single `str` or `python:list` of strings ['','']
    :param str OutputPath: Optional parameter to specify the Path in the Storage, postpended to /ilc/user/u/username/
    :param OutputSE: Optional parameter to specify the Storage Element to store data or files, e.g. CERN-SRM
    :type OutputSE: `python:list` or `str`
    """
    kwargs = {'lfns': lfns, 'OutputSE': OutputSE, 'OutputPath': OutputPath}
    if isinstance(lfns, list) and lfns:
      outputDataStr = ';'.join(lfns)
      description = 'List of output data files'
      self._addParameter(self.workflow, 'UserOutputData', 'JDL', outputDataStr, description)
    elif isinstance(lfns, six.string_types):
      description = 'Output data file'
      self._addParameter(self.workflow, 'UserOutputData', 'JDL', lfns, description)
    else:
      return self._reportError('Expected file name string or list of file names for output data', **kwargs)

    if OutputSE:
      description = 'User specified Output SE'
      if isinstance(OutputSE, six.string_types):
        OutputSE = [OutputSE]
      elif not isinstance(OutputSE, list):
        return self._reportError('Expected string or list for OutputSE', **kwargs)
      OutputSE = ';'.join(OutputSE)
      self._addParameter(self.workflow, 'UserOutputSE', 'JDL', OutputSE, description)

    if OutputPath:
      description = 'User specified Output Path'
      if not isinstance(OutputPath, six.string_types):
        return self._reportError('Expected string for OutputPath', **kwargs)
      # Remove leading "/" that might cause problems with os.path.join
      while OutputPath[0] == '/':
        OutputPath = OutputPath[1:]
      if OutputPath.count("ilc/user"):
        return self._reportError('Output path contains /ilc/user/ which is not what you want', **kwargs)
      self._addParameter(self.workflow, 'UserOutputPath', 'JDL', OutputPath, description)

    return S_OK()

  #############################################################################
  def setOutputSandbox(self, files):
    """Specify output sandbox files.  If specified files are over 10MB, these may be uploaded to Grid storage with a notification returned in the output sandbox.

    .. Note ::
       Sandbox files are removed after 2 weeks.

    Example usage:

    >>> job = UserJob()
    >>> job.setOutputSandbox(['*.log','*.sh', 'myfile.txt'])

    Use the output sandbox only for small files. Larger files should be stored
    on the grid and downloaded later if necessary. See :func:`setOutputData`

    :param files: Output sandbox files
    :type files: Single `str` or `python:list` of strings ['','']
    """
    if isinstance(files, list) and files:
      fileList = ";".join(files)
      description = 'Output sandbox file list'
      self._addParameter(self.workflow, 'OutputSandbox', 'JDL', fileList, description)
    elif isinstance(files, six.string_types):
      description = 'Output sandbox file'
      self._addParameter(self.workflow, 'OutputSandbox', 'JDL', files, description)
    else:
      kwargs = {'files': files}
      return self._reportError('Expected file string or list of files for output sandbox contents', **kwargs)

    return S_OK()

  def setILDConfig(self, version):
    """Define the Configuration package to obtain."""
    appName = 'ILDConfig'
    self._addSoftware(appName.lower(), version)
    self._addParameter(self.workflow, 'ILDConfigPackage', 'JDL', appName + version, 'ILDConfig package')
    return S_OK()



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
'''
:since: May 30, 2013

:author: sposs
'''

from __future__ import absolute_import
import os
import socket

from ILCDIRAC.Core.Utilities.CombinedSoftwareInstallation import getSharedAreaLocation
from ILCDIRAC.Workflow.Modules.ModuleBase import ModuleBase
from DIRAC.Core.Utilities.Os import getDiskSpace, getDirectorySize

from DIRAC import S_OK, S_ERROR, gLogger

__RCSID__ = '$Id$'
LOG = gLogger.getSubLogger(__name__)


class AnalyseWN(ModuleBase):
  """Module that will dump the host properties, and also the SharedArea content if it's there."""

  def __init__(self):
    """Constructor."""
    super(AnalyseWN, self).__init__()

  def execute(self):
    """Run the module."""
    result = self.resolveInputVariables()
    if not result['OK']:
      LOG.error("Failed to get the input parameters:", result['Message'])
      return result

    if not self.applicationLog:
      LOG.warn("Log file name missing, reverting to default")
      self.applicationLog = "AnalyseWN.log"

    info = []
    try:
      info.append("Host is %s" % socket.gethostname())
    except EnvironmentError:
      info.append("Could not determine host")

    size = getDiskSpace()
    if size > 0:
      info.append("Local disk is %s MB" % size)

    fileName = '/proc/cpuinfo'
    if os.path.exists(fileName):
      with open(fileName, 'r') as cpuInfoFile:
        cpu = cpuInfoFile.readlines()
      nCPU = 0
      for line in cpu:
        if line.find('cpu MHz') == 0:
          nCPU += 1
          freq = line.split()[3]
        elif line.find('model name') == 0:
          cpuModel = line.split(': ')[1].strip()
      info.append('CPU (model)    = %s' % cpuModel)
      info.append('CPU (MHz)      = %s x %s' % (nCPU, freq))

    fileName = '/proc/meminfo'
    if os.path.exists(fileName):
      with open(fileName, 'r') as memInfoFile:
        mem = memInfoFile.readlines()
      freeMem = 0
      for line in mem:
        if line.find('MemTotal:') == 0:
          totalMem = int(line.split()[1])
        if line.find('MemFree:') == 0:
          freeMem += int(line.split()[1])
        if line.find('Cached:') == 0:
          freeMem += int(line.split()[1])
      info.append('Memory (kB)    = %s' % totalMem)
      info.append('FreeMem. (kB)  = %s' % freeMem)

    fs = os.statvfs(".")
    # bsize;    /* file system block size */
    # frsize;   /* fragment size */
    # blocks;   /* size of fs in f_frsize units */
    # bfree;    /* # free blocks */
    # bavail;   /* # free blocks for non-root */
    # files;    /* # inodes */
    # ffree;    /* # free inodes */
    # favail;   /* # free inodes for non-root */
    # flag;     /* mount flags */
    # namemax;  /* maximum filename length */
    diskSpace = fs[4] * fs[0] // 1024 // 1024
    info.append('DiskSpace (MB) = %s' % diskSpace)

    sha = getSharedAreaLocation()
    if not sha:
      info.append("No shared Area found here")
    else:
      info.append("Shared Area found: %s" % sha)
      info.append("Content:")
      sha_list = os.listdir(sha)
      for item in sha_list:
        info.append("   %s" % item)
      sha_size = getDirectorySize(sha)
      if sha_size:
        info.append("It uses %s MB of disk" % sha_size)

    if os.path.isdir("/cvmfs/ilc.cern.ch"):
      info.append("Has CVMFS")

    try:
      of = open(self.applicationLog, "w")
      of.write("\n".join(info))
      of.close()
    except OSError:
      LOG.error("Could not create the log file")
      return S_ERROR("Failed saving the site info")

    return S_OK()



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Create productions for the Whizard2/DDSim/Marlin software chain.

First create a config file template::

  dirac-clic-make-productions -p

Then modify the template to describe the productions::

    [Production Parameters]
    detectorModel = CLIC_o3_v14
    version = 2018-08-10
    softwareVersion = ILCSoft-%(version)s_gcc62
    configVersion = ILCSoft-%(version)s
    generatorVersion = 2.6.3
    productionLogLevel = VERBOSE
    outputSE = CERN-DST-EOS
    finalOutputSE = CERN-SRM
    DRC=200
    machine = ee
    prodGroup = %(detectorModel)s_DiJets
    ProdTypes = Gen, Sim, Rec, RecOver
    energies =                100, 200
    processes =          DY_uds
    eventsPerJobs =           100
    MoveTypes = REC, GEN, SIM
    MoveStatus = Active
    MoveGroupSize = 10
    move = True
    overlayEvents = 380GeV
    cliReco = --Config.Tracking=Conformal --MyDDMarlinPandora.D0TrackCut=%(DRC)s
        --MyDDMarlinPandora.Z0TrackCut=%(DRC)s --MyDDMarlinPandora.MaxBarrelTrackerInnerRDistance=%(DRC)s
    additionalName = 3
    generatorSteeringFile = dy_uds.sin

Further options can be found in the created template file. Many options can contain comma separated values to submit
multiple chains of productions in one command. The example above will create two chains of Generation, Simulation,
Reconstruction and Reconstruction with overlay transformations, one for 100 GeV and one for 200 GeV.

Then test if everything works::

  dirac-clic-make-productions -f myProd

And finally submit to the system::

  dirac-clic-make-productions -f myProd -x

Options:

   -p, --printConfigFile      Create the template to create productions
   -f, --configFile <file>    Defines the file with the parameters to create a production
   -x, --enable               Disable dry-run mode and actually create the production
   --additionalName       Define a string to add to the production name if the original name already exists


Parameters in the steering file

  :configPackage: Steering file package to use for simulation and reconstruction
  :configVersion: Steering file version to use for simulation and reconstruction
  :detectorModel: Detector Model to use in simulation and reconstruction
  :outputSE: output SE for transformation jobs
  :finalOutputSE: final destination for files when moving transformations are enabled
  :machine: the name of the machine, ee, hh, eh
  :prodGroup: basename of the production group the productions are part of
  :productionLogLevel: log level to use in production jobs
  :softwareVersion: softwareVersion to use for generation/simulation/reconstruction
  :additionalName: additionalName to add to the transformation name in case a transformation
     with that name already exists

  :ProdTypes: Which transformations to create: Gen, Split, Sim, Rec, RecOver
  :MoveTypes: Which output file types to move: Gen, Sim, Rec, Dst
  :MoveGroupSize: The number of files to put in one replicationRequest
  :MoveStatus: The status of the Replication transformations: Active or Stopped
  :move: Whether or not to create the transformations to the output files to the finalOutputSE

  :energies: energy to use for generation or meta data search for each transformation chain
  :eventsPerJobs: number of events per job
  :processes: name of the processes to generate or use in meta data search
  :prodids: transformation IDs to use in meta data search for the first transformation of each chain

  :simSteeringFile: path to steering file for the generator
  :simVersion: specify which version of the generator to use
  :simApplication: specify which application to use for event generation

  :recoSteeringFile: path to steering file for the generator
  :recoVersion: specify which version of the generator to use
  :recoApplication: specify which application to use for event generation

  :generatorSteeringFile: path to steering file for the generator
  :generatorVersion: specify which version of the generator to use
  :generatorApplication: specify which application to use for event generation
  :numberOfTasks: number of production jobs/task to create for Gen transformations (default is 1)

  :eventsInSplitFiles: For split transformations, number of events in the input files

  :cliReco: additional CLI options for reconstruction

  :overlayEvents: For ``RecOver`` transformations use these events for Overlay. By default the gghad
     events with the process energy are used for overlay. E.g.: ``380GeV``


The individual applications can be further modified in their respective section::

  [Marlin]
  #ApplicationAttributeName=Value

  [DDSim]
  #ApplicationAttributeName=Value

  [Overlay]
  #ApplicationAttributeName=Value

  [Whizard2]
  #ApplicationAttributeName=Value

All attributes with a ``set`` method can be changed. See
:mod:`~ILCDIRAC.Interfaces.API.NewInterface.Applications.Whizard2`,
:mod:`~ILCDIRAC.Interfaces.API.NewInterface.Applications.Marlin`,
:mod:`~ILCDIRAC.Interfaces.API.NewInterface.Applications.OverlayInput`,
:mod:`~ILCDIRAC.Interfaces.API.NewInterface.Applications.DDSim`.



:since: July 14, 2017
:author: A Sailer
"""

# pylint disable=wrong-import-position

from __future__ import print_function
from __future__ import absolute_import
from pprint import pformat
from collections import defaultdict
from copy import deepcopy
from six.moves import range

import six.moves.configparser
import os

from DIRAC.Core.Base import Script
from DIRAC import S_OK, S_ERROR, gLogger

from ILCDIRAC.Core.Utilities.OverlayFiles import energyWithUnit, energyToInt
from ILCDIRAC.Core.Utilities.Utilities import listify, lowerFirst
from ILCDIRAC.ILCTransformationSystem.Utilities.Utilities import Task
import six
from six.moves import zip_longest

PRODUCTION_PARAMETERS = 'Production Parameters'
PP = PRODUCTION_PARAMETERS
APPLICATION_LIST = ['Marlin', 'DDSim', 'Overlay', 'Whizard2', 'KKMC']
LIST_ATTRIBUTES = ['ignoreMetadata',
                   'generatorSteeringFile',
                   'energies',
                   'eventsPerJobs',
                   'numberOfTasks',
                   'processes',
                   'prodIDs',
                   'eventsInSplitFiles',
                   'taskNames',
                   ]

STRING_ATTRIBUTES = ['configPackage',
                     'configVersion',
                     'additionalName',
                     'productionloglevel',
                     'outputSE',
                     'finalOutputSE',
                     'generatorVersion',
                     'generatorApplication',
                     'MoveStatus',
                     'MoveGroupSize',
                     'prodGroup',
                     'machine',
                     'detectorModel',
                     'softwareVersion',
                     'overlayEvents',
                     'overlayEventType',
                     ]


class _Params(object):
  """Parameter Object."""

  def __init__(self):
    self.prodConfigFilename = None
    self.dumpConfigFile = False
    self.dryRun = True
    self.additionalName = ''

  def setProdConf(self, fileName):
    if not os.path.exists(fileName):
      return S_ERROR("ERROR: File %r not found" % fileName)
    self.prodConfigFilename = fileName
    return S_OK()

  def setDumpConf(self, _):
    self.dumpConfigFile = True
    return S_OK()

  def setEnable(self, _):
    self.dryRun = False
    return S_OK()

  def setAddName(self, addName):
    self.additionalName = addName
    return S_OK()

  def registerSwitches(self):
    Script.registerSwitch("f:", "configFile=", "Set config file for production", self.setProdConf)
    Script.registerSwitch("x", "enable", "create productions, if off dry-run", self.setEnable)
    Script.registerSwitch("p", "printConfigFile", "Print a config file to stdout", self.setDumpConf)
    Script.registerSwitch("", "additionalName=", "Name to add to the production", self.setAddName)
    Script.setUsageMessage("""%s --configFile=myProduction""" % ("dirac-clic-make-productions", ))


class FCCDetProdChain(object):
  """Create applications and productions for CLIC physics studies."""

  class Flags(object):
    """flags to enable or disable productions.

    :param bool dryRun: if False no productions are created
    :param bool gen: if True create generation production
    :param bool spl: if True create split production
    :param bool sim: if True create simulation production
    :param bool rec: if True create reconstruction production
    :param bool over: if True create reconstruction production with overlay, if `rec` is False this flag is also False
    :param bool move: if True create moving transformations, the other move flags only take effect if this one is True
    :param bool moveGen: if True move GEN files after they have been used in the production, also for split files
    :param bool moveSim: if True move SIM files after they have been used in the production
    :param bool moveRev: if True move REC files when they were created
    :param bool moveDst: if True move DST files when they were created
    """

    def __init__(self):
      # general flag to create anything at all
      self._dryRun = True

      # create transformations
      self._gen = False
      self._spl = False
      self._sim = False
      self._rec = False
      self._over = False

      # create moving transformations
      self._moves = False
      self._moveGen = False
      self._moveSim = False
      self._moveRec = False
      self._moveDst = False

      # list of tuple to preserve order
      self._prodTypes = [('gen', 'Gen'), ('spl', 'Split'), ('sim', 'Sim'), ('rec', 'Rec'), ('over', 'RecOver')]
      self._moveTypes = [('moveGen', 'Gen'), ('moveSim', 'Sim'), ('moveRec', 'Rec'), ('moveDst', 'Dst')]

    @property
    def dryRun(self):  # pylint: disable=missing-docstring
      return self._dryRun

    @property
    def gen(self):  # pylint: disable=missing-docstring
      return self._gen

    @property
    def spl(self):  # pylint: disable=missing-docstring
      return self._spl

    @property
    def sim(self):  # pylint: disable=missing-docstring
      return self._sim

    @property
    def rec(self):  # pylint: disable=missing-docstring
      return self._rec

    @property
    def over(self):  # pylint: disable=missing-docstring
      return self._over

    @property
    def move(self):  # pylint: disable=missing-docstring
      return self._moves

    @property
    def moveGen(self):  # pylint: disable=missing-docstring
      return (self._gen or self._spl) and self._moves and self._moveGen

    @property
    def moveSim(self):  # pylint: disable=missing-docstring
      return self._sim and self._moves and self._moveSim

    @property
    def moveRec(self):  # pylint: disable=missing-docstring
      return (self._rec or self._over) and self._moves and self._moveRec

    @property
    def moveDst(self):  # pylint: disable=missing-docstring
      return (self._rec or self._over) and self._moves and self._moveDst

    def __str__(self):
      pDict = vars(self)
      self.updateDictWithFlags(pDict)
      return """

#Productions to create: %(prodOpts)s
ProdTypes = %(prodTypes)s

move = %(_moves)s

#Datatypes to move: %(moveOpts)s
MoveTypes = %(moveTypes)s
""" % (vars(self))

    def updateDictWithFlags(self, pDict):
      """add flags and values to pDict."""
      for attr in dir(self):
        if isinstance(getattr(type(self), attr, None), property):
          pDict.update({attr: str(getattr(self, attr))})

      pDict.update(prodOpts=", ".join([pTuple[1]
                                           for pTuple in self._prodTypes]))
      pDict.update(prodTypes=", ".join([pTuple[1]
                                            for pTuple in self._prodTypes
                                            if getattr(self, pTuple[0])]))
      pDict.update(moveOpts=", ".join([pTuple[1]
                                            for pTuple in self._moveTypes]))
      pDict.update(moveTypes=", ".join([pTuple[1]
                                            for pTuple in self._moveTypes
                                            if getattr(self, '_' + pTuple[0])]))

    def __splitStringToOptions(self, config, tuples, optString, prefix='_'):
      """split the option string into separate values and set the corresponding flag."""
      prodsToCreate = config.get(PRODUCTION_PARAMETERS, optString)
      for prodType in listify(prodsToCreate):
        if not prodType:
          continue
        found = False
        for attribute, name in tuples:
          if name.capitalize() == prodType.strip().capitalize():
            setattr(self, prefix + attribute, True)
            found = True
            break
        if not found:
          raise AttributeError("Unknown parameter: %r " % prodType)

    def loadFlags(self, config):
      """load flags values from configfile."""
      self.__splitStringToOptions(config, self._prodTypes, 'ProdTypes', prefix='_')
      self.__splitStringToOptions(config, self._moveTypes, 'MoveTypes', prefix='_')
      self._moves = config.getboolean(PP, 'move')

  def __init__(self, params=None, group='ilc_prod'):

    from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
    self._ops = Operations(group=group)

    self.machine = {'ilc_prod': 'clic',
                    'fcc_prod': 'fccee',
                    }[group]

    self.overlayEventType = {'ilc_prod': 'gghad',
                             'fcc_prod': 'pairs',
                             }[group]

    self.prodGroup = 'several'
    self._basepath = None  # resolved lazily via property
    prodPath = os.path.join('/Production', self.machine)
    self.detectorModel = self._ops.getValue(os.path.join(prodPath, 'DefaultDetectorModel'))
    self.softwareVersion = self._ops.getValue(os.path.join(prodPath, 'DefaultSoftwareVersion'))
    self.configVersion = self._ops.getValue(os.path.join(prodPath, 'DefaultConfigVersion'))
    self.configPackage = self._ops.getValue(os.path.join(prodPath, 'DefaultConfigPackage'))
    self.productionLogLevel = 'VERBOSE'
    self.outputSE = 'CERN-DST-EOS'
    self.moveStatus = 'Stopped'
    self.moveGroupSize = '10'

    self.eventsPerJobs = []
    self.numberOfTasks = []
    self.energies = []
    self.processes = []
    self.prodIDs = []
    self.eventsInSplitFiles = []
    self.taskNames = []

    # final destination for files once they have been used
    self.finalOutputSE = self._ops.getValue('Production/CLIC/FailOverSE')

    self.additionalName = params.additionalName

    self.overlayEvents = ''

    self.cliRecoOption = ''
    self.cliReco = ''

    self.generatorApplication = 'whizard2'
    self.generatorVersion = None
    self.generatorSteeringFile = []

    self.simApplication = 'ddsim'
    self.simVersion = None
    self.simSteeringFile = 'clic_steer.py'

    self.recoApplication = 'Marlin'
    self.recoVersion = None
    self.recoSteeringFile = 'clicReconstruction.xml'


    self.ignoreMetadata = []

    self.applicationOptions = {appName: {} for appName in APPLICATION_LIST}

    self._flags = self.Flags()

    self.loadParameters(params)

    self._flags._dryRun = params.dryRun  # pylint: disable=protected-access

  @property
  def basepath(self):
    """Return the basepath"""
    if not self._basepath:
      prodPath = os.path.join('/Production', self.machine.upper())
      self._basepath = self._ops.getValue(os.path.join(prodPath, 'BasePath'))
    return self._basepath
  
  def meta(self, prodID, process, energy):
    """return meta data dictionary, always new."""
    metaD = {'ProdID': str(prodID),
             'EvtType': process,
             'Energy': self.metaEnergy(energy),
             'Machine': self.machine,
             }
    for key in self.ignoreMetadata:
      metaD.pop(key)
    return metaD

  def loadParameters(self, parameter):
    """Load parameters from config file."""
    if parameter.prodConfigFilename is not None:
      defaultValueDict = vars(self)
      self._flags.updateDictWithFlags(defaultValueDict)
      # we are passing all instance attributes as the default dict so generally we do not have to check
      # if an option exists, also options are case insensitive and stored in lowercase
      config = six.moves.configparser.SafeConfigParser(defaults=defaultValueDict, dict_type=dict)
      config.read(parameter.prodConfigFilename)
      self._flags.loadFlags(config)

      for attribute in LIST_ATTRIBUTES:
        setattr(self, attribute, listify(config.get(PP, attribute)))

      for attribute in STRING_ATTRIBUTES:
        setattr(self, lowerFirst(attribute), config.get(PP, attribute))

      # this parameter is deprecated and not part of the instance attributes so we need to check for existence
      if config.has_option(PP, 'clicConfig'):
        gLogger.warn('"clicConfig" parameter is deprected, please dump a new steering file!')
        self.configVersion = config.get(PP, 'clicConfig')

      # attribute and option names differ, special treatment
      self.cliRecoOption = config.get(PP, 'cliReco')

      if self.moveStatus not in ('Active', 'Stopped'):
        raise AttributeError("MoveStatus can only be 'Active' or 'Stopped' not %r" % self.moveStatus)

      # self.overlayEvents = self.checkOverlayParameter(self.overlayEvents)
      # self.overlayEventType = self.overlayEventType + self.overlayEvents.lower()

      self.processes = [process.strip() for process in self.processes if process.strip()]
      self.energies = [float(eng.strip()) for eng in self.energies if eng.strip()]
      self.eventsPerJobs = [int(epj.strip()) for epj in self.eventsPerJobs if epj.strip()]
      # these do not have to exist so we fill them to the same length if they are not set
      self.prodIDs = [int(pID.strip()) for pID in self.prodIDs if pID.strip()]
      self.prodIDs = self.prodIDs if self.prodIDs else [1 for _ in self.energies]

      # if one of the lists only has size 1 and there is a longer list we extend
      # the list to the maximum size assuming the values are re-used
      maxLength = 0
      parameterLists = [self.processes, self.energies, self.eventsPerJobs, self.generatorSteeringFile]
      for parList in parameterLists:
        maxLength = len(parList) if len(parList) > maxLength else maxLength
      for parList in parameterLists:
        if len(parList) == 1 and maxLength > 1:
          parList.extend([parList[0]] * (maxLength - 1))

      if not (self.processes and self.energies and self.eventsPerJobs) and self.prodIDs:
        eventsPerJobSave = list(self.eventsPerJobs) if self.eventsPerJobs else None
        self._getProdInfoFromIDs()
        self.eventsPerJobs = eventsPerJobSave if eventsPerJobSave else self.eventsPerJobs

      self.numberOfTasks = [int(nbtask.strip()) for nbtask in self.numberOfTasks if nbtask.strip()]
      self.numberOfTasks = self.numberOfTasks if self.numberOfTasks else [1] * len(self.energies)
      self.taskNames = self.taskNames if self.taskNames else [''] * len(self.energies)

      if len(self.processes) != len(self.energies) or \
         len(self.energies) != len(self.eventsPerJobs) or \
         len(self.prodIDs) != len(self.eventsPerJobs) or \
         len(self.eventsPerJobs) != len(self.taskNames) or \
         False:
        raise AttributeError('Lengths of Processes, Energies, EventsPerJobs, AdditionalNames do not match')

      if self._flags.gen:
        if len(self.numberOfTasks) != len(self.energies) or \
           (self.generatorSteeringFile and (len(self.generatorSteeringFile) != len(self.energies))):
          raise AttributeError("Lengths of numberOfTasks, generatorSteeringFile, and Energies do not match")

      self.eventsInSplitFiles = listify(self.eventsInSplitFiles, int)
      self.eventsInSplitFiles = self.eventsInSplitFiles if self.eventsInSplitFiles else [-1] * len(self.energies)

      if self._flags.spl and len(self.eventsInSplitFiles) != len(self.energies):
        raise AttributeError("Length of eventsInSplitFiles does not match: %d vs %d" % (
            len(self.eventsInSplitFiles),
            len(self.energies)))

      # read options from application sections
      config2 = six.moves.configparser.SafeConfigParser(dict_type=dict)
      config2.optionxform = str  # do not transform options to lowercase
      config2.read(parameter.prodConfigFilename)
      for appName in APPLICATION_LIST:
        try:
          self.applicationOptions[appName] = dict(config2.items(appName))
        except six.moves.configparser.NoSectionError:
          pass

    if parameter.dumpConfigFile:
      print(self)
      raise RuntimeError('')

  def _getProdInfoFromIDs(self):
    """get the processName, energy and eventsPerJob from the MetaData catalog.

    :raises: AttributeError if some of the information cannot be found
    :returns: None
    """
    if not self.prodIDs:
      raise AttributeError("No prodIDs defined")

    self.eventsPerJobs = []
    self.processes = []
    self.energies = []
    from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
    from DIRAC.Resources.Catalog.FileCatalogClient import FileCatalogClient
    trc = TransformationClient()
    fc = FileCatalogClient()
    for prodID in self.prodIDs:
      gLogger.notice("Getting information for %s" % prodID)
      tRes = trc.getTransformation(str(prodID))
      if not tRes['OK']:
        raise AttributeError("No prodInfo found for %s" % prodID)
      self.eventsPerJobs.append(int(tRes['Value']['EventsPerTask']))
      lfnRes = fc.findFilesByMetadata({'ProdID': prodID})
      if not lfnRes['OK'] or not lfnRes['Value']:
        raise AttributeError("Could not find files for %s: %s " % (prodID, lfnRes.get('Message', lfnRes.get('Value'))))
      path = os.path.dirname(lfnRes['Value'][0])
      fileRes = fc.getDirectoryUserMetadata(path)
      self.processes.append(fileRes['Value']['EvtType'])
      self.energies.append(fileRes['Value']['Energy'])
      gLogger.notice("Found (Evts,Type,Energy): %s %s %s " %
                     (self.eventsPerJobs[-1], self.processes[-1], self.energies[-1]))

  def __str__(self):
    pDict = vars(self)
    appOptionString = ''
    for appName in APPLICATION_LIST:
      appOptionString += '[%s]\n#ApplicationAttributeName=Value\n\n' % appName

    pDict.update({'ProductionParameters': PRODUCTION_PARAMETERS})
    pDict.update({'ApplicationOptions': appOptionString})
    return """
%(ApplicationOptions)s
[%(ProductionParameters)s]
machine = %(machine)s
prodGroup = %(prodGroup)s
detectorModel = %(detectorModel)s
softwareVersion = %(softwareVersion)s

generatorApplication = %(generatorApplication)s
generatorVersion = %(generatorVersion)s
generatorSteeringFile = %(generatorSteeringFile)s

simApplication = %(simApplication)s
simVersion = %(simVersion)s
simSteeringFile = %(simSteeringFile)s

recoApplication = %(recoApplication)s
recoVersion = %(recoVersion)s
recoSteeringFile = %(recoSteeringFile)s

configVersion = %(configVersion)s
configPackage = %(configPackage)s
eventsPerJobs = %(eventsPerJobs)s
## Number of jobs/task to generate (default = 1)
# numberOfTasks =

energies = %(energies)s
processes = %(processes)s
## optional prodid to search for input files
# prodIDs =

## number of events for input files to split productions
eventsInSplitFiles = %(eventsInSplitFiles)s

productionLogLevel = %(productionLogLevel)s
outputSE = %(outputSE)s

finalOutputSE = %(finalOutputSE)s
MoveStatus = %(moveStatus)s
MoveGroupSize = %(moveGroupSize)s

## optional additional name
# additionalName = %(additionalName)s
## optional additional names, for for each process, prodID, etc.
# taskNames =

## optional marlin CLI options
# cliReco = %(cliReco)s

overlayEventType = %(overlayEventType)s
## optional energy to use for overlay: e.g. 3TeV
# overlayEvents = %(overlayEvents)s

%(_flags)s

""" % (pDict)

  @staticmethod
  def metaEnergy(energy):
    """Return string of the energy with no non-zero digits."""
    if isinstance(energy, six.string_types):
      return energy
    energy = ("%1.2f" % energy).rstrip('0').rstrip('.')
    return energy

  def createGeneratorApplication(self, task):
    """Create the selected generator application."""

    genApp =  {'kkmc': self.createKKMCApplication,
             }[self.generatorApplication.lower()](task)
    if task.datatype:
      genApp.datatype = task.datatype
    return genApp


  def createSimulationApplication(self, task):
    """Create the selected Simulation application."""

    return {'genericapp': self.createGenericApplication,
            }[self.simApplication.lower()](task)

  def createKKMCApplication(self, task):
    """create KKMCee Application."""
    from ILCDIRAC.Interfaces.API.NewInterface.Applications import KKMC

    kkmcee = KKMC()
    kkmcee.setVersion(self.generatorVersion)
    if task.genFile:
      kkmcee.setConfigFile(task.genFile)
    kkmcee.setEvtType(''.join(task.meta['EvtType'][:len(task.meta['EvtType']) / 2]).capitalize())
    kkmcee.setNumberOfEvents(task.eventsPerJob)
    kkmcee.setEnergy(task.meta['Energy'])
    self._setApplicationOptions('KKMC', kkmcee, task.applicationOptions)
    task.datatype = 'lhef'
    task.generator = 'kkmcee'
    return kkmcee

  def createGenericApplication(self, task):
    """Create GenericApplication."""
    from ILCDIRAC.Interfaces.API.NewInterface.Applications import GaudiApp

    ga = GaudiApp()
    ga.setVersion('key4hep-latest')  # FIXME: task.gaSetupScript
    ga.setExecutable("DelphesPythia8_EDM4HEP")  # FIXME task.gaScript
    ga.setInputFileFlag('')
    # FIXME this should be configurable
    ga.setOutputFileFlag("delphes_card_IDEA.tcl edm4hep_output_config.tcl Pythia_LHEinput.cmd")
    # task.gaArguments
    # ga.setEvtType(task.meta['EvtType'])
    ga.setNumberOfEvents(task.eventsPerJob)
    ga.setEnergy(task.meta['Energy'])
    ga._extension = 'root'
    ga.datatype = 'delphes'
    self._setApplicationOptions('GA', ga, task.applicationOptions)

    return ga

  def createGenerationProduction(self, task):
    """Create generation production."""
    prodName = task.getProdName(self.machine, 'gen', self.additionalName)
    parameterDict = task.parameterDict
    nbTasks = task.nbTasks
    gLogger.notice("*" * 80 + "\nCreating generation production: %s " % prodName)
    genProd = self.getProductionJob()
    genProd.setConfigPackage(appName=self.configPackage, version=self.configVersion)
    genProd.evttype = parameterDict['process']
    genProd.setProdType('MCGeneration')
    genProd.setWorkflowName(prodName)
    # this might modify task
    genApp = self.createGeneratorApplication(task)
    if task.generator:
      genProd.generator = task.generator
    res = genProd.append(genApp)
    if not res['OK']:
      raise RuntimeError("Error creating generation production: %s" % res['Message'])

    if True:  # delphes after the KKMC task
      gaApp = self.createGenericApplication(task)
      gaApp.detector = 'IDEA'
      gaApp.datatype = 'delphes'
      res = genProd.append(gaApp)
      if not res['OK']:
        raise RuntimeError("Error creating generation production: %s" % res['Message'])

    genProd.addFinalization(True, True, True, True)
    if not prodName:
      raise RuntimeError("Error creating generation production: prodName empty")
    genProd.setDescription(prodName)
    res = genProd.createProduction()
    if not res['OK']:
      raise RuntimeError("Error creating generation production: %s" % res['Message'])

    # genProd.addMetadataToFinalFiles({'BeamParticle1': parameterDict['pname1'],
    #                                  'BeamParticle2': parameterDict['pname2'],
    #                                  'EPA_B1': parameterDict['epa_b1'],
    #                                  'EPA_B2': parameterDict['epa_b2'],
    #                                 }
    #                                )

    res = genProd.finalizeProd()
    if not res['OK']:
      raise RuntimeError("Error finalizing generation production: %s" % res['Message'])

    genProd.setNbOfTasks(nbTasks)
    generationMeta = genProd.getMetadata()
    return generationMeta

  def createSimulationProduction(self, task):
    """Create simulation production."""
    meta = task.meta
    prodName = task.getProdName('sim', self.detectorModel, self.additionalName)
    parameterDict = task.parameterDict
    gLogger.notice("*" * 80 + "\nCreating simulation production: %s " % prodName)
    simProd = self.getProductionJob()
    simProd.setProdType('MCSimulation')
    simProd.setConfigPackage(appName=self.configPackage, version=self.configVersion)
    res = simProd.setInputDataQuery(meta)
    if not res['OK']:
      raise RuntimeError("Error creating Simulation Production: %s" % res['Message'])
    simProd.setWorkflowName(prodName)
    # Add the application
    res = simProd.append(self.createSimulationApplication(task))
    if not res['OK']:
      raise RuntimeError("Error creating simulation Production: %s" % res['Message'])
    simProd.addFinalization(True, True, True, True)
    description = "Model: %s" % self.detectorModel
    if prodName:
      description += ", %s" % prodName
    simProd.setDescription(description)
    res = simProd.createProduction()
    if not res['OK']:
      raise RuntimeError("Error creating simulation production: %s" % res['Message'])

    simProd.addMetadataToFinalFiles({'BeamParticle1': parameterDict['pname1'],
                                       'BeamParticle2': parameterDict['pname2'],
                                       'EPA_B1': parameterDict['epa_b1'],
                                       'EPA_B2': parameterDict['epa_b2'],
                                     }
                                   )

    res = simProd.finalizeProd()
    if not res['OK']:
      raise RuntimeError("Error finalizing simulation production: %s" % res['Message'])

    simulationMeta = simProd.getMetadata()
    return simulationMeta

  def createReconstructionProduction(self, task, over):
    """Create reconstruction production."""
    gLogger.notice('Creating Reconstruction production', task)
    meta = task.meta
    recType = 'rec_overlay' if over else 'rec'
    prodName = task.getProdName(recType, self.detectorModel, self.additionalName)
    if over:
      prodName = prodName.replace('overlay', 'overlay%s' % self.overlayEvents if self.overlayEvents else meta['Energy'])
    parameterDict = task.parameterDict
    gLogger.notice("*" * 80 + "\nCreating %s reconstruction production: %s " % ('overlay' if over else '', prodName))
    recProd = self.getProductionJob()
    productionType = 'MCReconstruction_Overlay' if over else 'MCReconstruction'
    recProd.setProdType(productionType)
    recProd.setConfigPackage(appName=self.configPackage, version=self.configVersion)

    res = recProd.setInputDataQuery(meta)
    if not res['OK']:
      raise RuntimeError("Error setting inputDataQuery for Reconstruction production: %s " % res['Message'])

    recProd.setWorkflowName(prodName)

    # # Add overlay if needed
    # if over:
    #   res = recProd.append(self.createOverlayApplication(task))
    #   if not res['OK']:
    #     raise RuntimeError("Error appending overlay to reconstruction transformation: %s" % res['Message'])

    # # add reconstruction
    # res = recProd.append(self.createMarlinApplication(task, over))
    # if not res['OK']:
    #   raise RuntimeError("Error appending Marlin to reconstruction production: %s" % res['Message'])
    # recProd.addFinalization(True, True, True, True)

    description = "CLICDet2017 %s" % meta['Energy']
    description += "Overlay" if over else "No Overlay"
    if prodName:
      description += ", %s" % prodName
    recProd.setDescription(description)

    res = recProd.createProduction()
    if not res['OK']:
      raise RuntimeError("Error creating reconstruction production: %s" % res['Message'])

    recProd.addMetadataToFinalFiles({'BeamParticle1': parameterDict['pname1'],
                                       'BeamParticle2': parameterDict['pname2'],
                                       'EPA_B1': parameterDict['epa_b1'],
                                       'EPA_B2': parameterDict['epa_b2'],
                                     }
                                   )

    res = recProd.finalizeProd()
    if not res['OK']:
      raise RuntimeError("Error finalising reconstruction production: %s " % res['Message'])

    reconstructionMeta = recProd.getMetadata()
    return reconstructionMeta

  def createSplitProduction(self, task, limited=False):
    """Create splitting transformation for splitting files."""
    meta = task.meta
    prodName = task.getProdName('split', task.meta['ProdID'], self.additionalName)
    parameterDict = task.parameterDict
    eventsPerJob = task.eventsPerJob
    eventsPerBaseFile = task.eventsPerBaseFile

    gLogger.notice("*" * 80 + "\nCreating split production: %s " % prodName)
    splitProd = self.getProductionJob()
    splitProd.setProdPlugin('Limited' if limited else 'Standard')
    splitProd.setProdType('Split')

    res = splitProd.setInputDataQuery(meta)
    if not res['OK']:
      raise RuntimeError('Split production: failed to set inputDataQuery: %s' % res['Message'])
    splitProd.setWorkflowName(prodName)

    # # Add the application
    # res = splitProd.append(self.createSplitApplication(eventsPerJob, eventsPerBaseFile, 'stdhep'))
    # if not res['OK']:
    #   raise RuntimeError('Split production: failed to append application: %s' % res['Message'])
    # splitProd.addFinalization(True, True, True, True)
    # description = 'Splitting stdhep files'
    # splitProd.setDescription(description)

    res = splitProd.createProduction()
    if not res['OK']:
      raise RuntimeError("Failed to create split production: %s " % res['Message'])

    splitProd.addMetadataToFinalFiles({"BeamParticle1": parameterDict['pname1'],
                                         "BeamParticle2": parameterDict['pname2'],
                                         "EPA_B1": parameterDict['epa_b1'],
                                         "EPA_B2": parameterDict['epa_b2'],
                                       }
                                     )

    res = splitProd.finalizeProd()
    if not res['OK']:
      raise RuntimeError('Split production: failed to finalize: %s' % res['Message'])

    return splitProd.getMetadata()

  def createMovingTransformation(self, meta, prodType):
    """create moving transformations for output files."""

    sourceSE = self.outputSE
    targetSE = self.finalOutputSE
    prodID = meta['ProdID']
    try:
      dataTypes = {'MCReconstruction': ('DST', 'REC'),
                    'MCReconstruction_Overlay': ('DST', 'REC'),
                    'MCSimulation': ('SIM',),
                    'MCGeneration': ('GEN',),
                  }[prodType]
    except KeyError:
      raise RuntimeError("ERROR creating MovingTransformation" + repr(prodType) + "unknown")

    if not any(getattr(self._flags, "move%s" % dataType.capitalize()) for dataType in dataTypes):
      gLogger.notice("*" * 80 + "\nNot creating moving transformation for prodID: %s, %s " % (meta['ProdID'], prodType))
      return

    from DIRAC.TransformationSystem.Utilities.ReplicationTransformation import createDataTransformation
    from DIRAC.TransformationSystem.Client.Transformation import Transformation
    for dataType in dataTypes:
      if getattr(self._flags, "move%s" % dataType.capitalize()):
        gLogger.notice("*" * 80 + "\nCreating moving transformation for prodID: %s, %s, %s " %
                       (meta['ProdID'], prodType, dataType))
        parDict = dict(flavour='Moving',
                       targetSE=targetSE,
                       sourceSE=sourceSE,
                       plugin='Broadcast%s' % ('' if dataType.lower() not in ('gen', 'sim') else 'Processed'),
                       metaKey='ProdID',
                       metaValue=prodID,
                       extraData={'Datatype': dataType},
                       tGroup=self.prodGroup,
                       groupSize=int(self.moveGroupSize),
                       enable=not self._flags.dryRun,
                      )
        message = "Moving transformation with parameters"
        gLogger.notice("%s:\n%s" % (message, pformat(parDict, indent=len(message) + 2, width=120)))
        res = createDataTransformation(**parDict)
        if not res['OK']:
          gLogger.error("Failed to create moving transformation:", res['Message'])

        elif isinstance(res['Value'], Transformation):
          newTrans = res['Value']
          newTrans.setStatus(self.moveStatus)

  def getProductionJob(self):
    """return production job instance with some parameters set."""
    from ILCDIRAC.Interfaces.API.NewInterface.ProductionJob import ProductionJob
    prodJob = ProductionJob()
    prodJob.setLogLevel(self.productionLogLevel)
    prodJob.setProdGroup(self.prodGroup)
    prodJob.setOutputSE(self.outputSE)
    prodJob.basepath = self.basepath
    prodJob.dryrun = self._flags.dryRun
    prodJob.maxFCFoldersToCheck = 1
    return prodJob

  def _setApplicationOptions(self, appName, app, optionsDict=None):
    """set options for given application.

    :param str appName: name of the application, for print out
    :param app: application instance
    """
    if optionsDict is None:
      optionsDict = {}
    allOptions = dict(self.applicationOptions.get(appName, {}))
    allOptions.update(optionsDict)
    for option, value in allOptions.items():
      if option.startswith(('FE.', 'C_', 'additionalName')):
        continue
      gLogger.notice("%s: setting option %s to %s" % (appName, option, value))
      setterFunc = 'set' + option
      if not hasattr(app, setterFunc):
        raise AttributeError("Cannot set %s for %s, check spelling!" % (option, appName))
      if value.lower() in ('false', 'true'):
        value = value.lower() == 'true'
      getattr(app, setterFunc)(value)

  def createTransformations(self, taskDict):
    """Create all the transformations we want to create."""
    for pType, createProduction in [('GEN', self.createGenerationProduction),
                                    ('SPLIT', self.createSplitProduction)]:
      for task in taskDict.get(pType, []):
        meta = createProduction(task)
        self.addSimTask(taskDict, meta, originalTask=task)
        taskDict['MOVE_' + pType].append(dict(meta))

    for task in taskDict.get('SIM', []):
      if not self._flags.sim:
        continue
      gLogger.notice("Creating task %s" % task)
      simMeta = self.createSimulationProduction(task)
      self.addRecTask(taskDict, simMeta, originalTask=task)
      taskDict['MOVE_SIM'].append(dict(simMeta))

    for task in taskDict.get('REC', []):
      for name, over, enabled in [('REC', False, self._flags.rec),
                                  ('OVER', True, self._flags.over)]:
        if enabled:
          recMeta = self.createReconstructionProduction(task, over=over)
          taskDict['MOVE_' + name].append(dict(recMeta))

    for name, pType in [('GEN', 'MCGeneration'),
                        ('SPLIT', 'MCGeneration'),
                        ('SIM', 'MCSimulation'),
                        ('REC', 'MCReconstruction'),
                        ('OVER', 'MCReconstruction_Overlay')]:
      for meta in taskDict.get('MOVE_' + name, []):
        self.createMovingTransformation(meta, pType)

  def createTaskDict(self, prodID, process, energy, eventsPerJob, sinFile, nbTasks,
                     eventsPerBaseFile, taskName):
    """Create a dictionary of tasks for the first level of transformations."""
    taskDict = defaultdict(list)
    metaInput = self.meta(prodID, process, energy)
    prodName = metaInput['EvtType']
    parameterDict = {'process': metaInput['EvtType']}
    if self._flags.gen:
      self.addGenTask(taskDict, Task(metaInput, parameterDict, eventsPerJob, nbTasks=nbTasks,
                                     genFile=sinFile,
                                     taskName=taskName))

    elif self._flags.spl and eventsPerBaseFile == eventsPerJob:
      gLogger.notice("*" * 80 + "\nSkipping split transformation for %s\n" % prodName + "*" * 80)
      if self._flags.sim:
        self.addSimTask(taskDict, metaInput, Task({}, parameterDict, eventsPerJob,
                                                  taskName=taskName))
    elif self._flags.spl:
      taskDict['SPLIT'].append(Task(metaInput, parameterDict, eventsPerJob,
                                    eventsPerBaseFile=eventsPerBaseFile))
    elif self._flags.sim:
      self.addSimTask(taskDict, metaInput, Task({}, parameterDict, eventsPerJob, taskName=taskName))
    elif self._flags.rec or self._flags.over:
      self.addRecTask(taskDict, metaInput, Task({}, parameterDict, eventsPerJob, taskName=taskName))

    return taskDict

  def _addTask(self, taskDict, metaInput, originalTask, prodType, applicationName):
    """Add a task to the taskDict."""
    options = defaultdict(list)
    nTasks = 0
    for option, value in self.applicationOptions[applicationName].items():
      if option.startswith('FE.'):
        optionName = option.split('.', 1)[1]
        options[optionName] = listify(value)
        gLogger.notice("Found option %s with values %s" % (optionName, pformat(options[optionName])))
        nTasks = len(options[optionName])

    theTask = Task(metaInput,
                   parameterDict=originalTask.parameterDict,
                   eventsPerJob=originalTask.eventsPerJob,
                   metaPrev=originalTask.meta,
                   genFile=originalTask.genFile,
                   nbTasks=originalTask.nbTasks,
                   )
    theTask.sourceName = '_'.join([originalTask.sourceName, originalTask.taskName])
    if not nTasks:
      taskDict[prodType].append(theTask)
      return

    taskList = [deepcopy(theTask) for _ in range(nTasks)]
    taskDict[prodType].extend(taskList)
    self.addTaskOptions(options, taskList)
    return

  def addGenTask(self, taskDict, originalTask):
    """Add a gen task with required options."""
    return self._addTask(taskDict, metaInput=originalTask.meta,
                         originalTask=originalTask, prodType='GEN',
                         applicationName=self.generatorApplication,
                         )

  def addRecTask(self, taskDict, metaInput, originalTask):
    """Add a reconstruction task."""
    return self._addTask(taskDict, metaInput, originalTask, prodType='REC', applicationName='Marlin')

  def addSimTask(self, taskDict, metaInput, originalTask):
    """Add a sim task."""
    return self._addTask(taskDict, metaInput, originalTask, prodType='SIM', applicationName='DDSim')

  @staticmethod
  def addTaskOptions(options, taskList):
    """Add the options to each task in the taskList."""
    for optionName, values in options.items():
      if optionName.startswith('Query'):
        queryParameter = optionName[len('Query'):]
        for index, value in enumerate(values):
          taskList[index].meta[queryParameter] = value
      elif optionName == 'additionalName':
        for index, value in enumerate(values):
          taskList[index].taskName = value
      # cliReco only makes sense for REC application, but it is otherwise ignored
      elif optionName == 'cliReco':
        for index, value in enumerate(values):
          taskList[index].cliReco = value
      else:
        for index, value in enumerate(values):
          taskList[index].applicationOptions[optionName] = value

  def createAllTransformations(self):
    """Loop over the list of processes, energies and possibly prodIDs to create all the productions."""
    for energy, process, prodID, eventsPerJob, eventsPerBaseFile, genFile, nbTasks, taskName in \
        zip_longest(self.energies, self.processes, self.prodIDs, self.eventsPerJobs, self.eventsInSplitFiles,
                     self.generatorSteeringFile, self.numberOfTasks, self.taskNames, fillvalue=None):
      taskDict = self.createTaskDict(prodID, process, energy, eventsPerJob, genFile,
                                     nbTasks, eventsPerBaseFile, taskName)
      self.createTransformations(taskDict)


if __name__ == "__main__":
  CLIP = _Params()
  CLIP.registerSwitches()
  Script.parseCommandLine()
  from ILCDIRAC.Core.Utilities.CheckAndGetProdProxy import checkOrGetGroupProxy
  CHECKGROUP = checkOrGetGroupProxy(['ilc_prod', 'fcc_prod'])
  if CHECKGROUP['OK']:
    pass
  elif CLIP.dryRun:
    gLogger.notice('Did not find correct group, dryRun enabled, assuming "ilc_prod"')
    CHECKGROUP = S_OK('ilc_prod')
  else:
    exit(1)
  try:
    CHAIN = FCCDetProdChain(params=CLIP, group=CHECKGROUP['Value'])
    CHAIN.createAllTransformations()
  except (AttributeError, RuntimeError) as excp:
    if str(excp) != '':
      gLogger.exception('Failure to create transformations', lException=excp)
      exit(1)
  exit(0)



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Gaudi Application to run applications based on Gaudi.

.. versionadded:: v32r0p1

Usage:

>>> ga.setVersion('key4hep_nightly')
>>> ga.setExecutable('k4run')
>>> ga.setSteeringFile('geant_fullsim_fccee_lar_pgun.py')
>>> ga.setDetectorModel("DetFCCeeIDEA-LAr")
>>> ga.setArguments("--GenAlg.MomentumRangeParticleGun.MomentumMin 1000 "
>>>                 "--GenAlg.MomentumRangeParticleGun.MomentumMax 1000 "
>>>                 "--GenAlg.MomentumRangeParticleGun.PdgCodes 11")
>>> ga.setOutputFile("output.root")

"""
from __future__ import absolute_import
import types
import os

from ILCDIRAC.Interfaces.API.NewInterface.LCApplication import LCApplication
from ILCDIRAC.Core.Utilities.InstalledFiles import Exists
from ILCDIRAC.Interfaces.Utilities.DDInterfaceMixin import DDInterfaceMixin
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Workflow.Parameter import Parameter
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
import six

LOG = gLogger.getSubLogger(__name__)
__RCSID__ = "$Id$"


class GaudiApp(DDInterfaceMixin, LCApplication):
  """GaudiApp Application Class."""

  def __init__(self, paramdict=None):
    self.startFrom = 0
    self.randomSeed = -1
    self.detectorModel = ''
    self.executableName = 'k4run'
    self.inputFileFlag = '--EventDataSvc.inputs'
    self.outputFileFlag = '--out.filename'
    self.detectorModelFlag = '--GeoSvc.detectors'
    super(GaudiApp, self).__init__(paramdict)
    # Those 5 need to come after default constructor
    self._modulename = 'GaudiAppModule'
    self._moduledescription = 'Module to run GaudiApp'
    self.appname = 'gaudiapp'
    self._ops = Operations()

  @property
  def detectortype(self):
    """DetectorType needed for transformations.

    Backward compatibility needed for Mokka?
    """
    return self.detectorModel

  @detectortype.setter
  def detectortype(self, value):
    """Ignore setting of detector type for GaudiApp."""
    pass

  def setRandomSeed(self, randomSeed):
    """Define random seed to use. Default is the jobID.

    :param int randomSeed: Seed to use during simulation.
    """
    self._checkArgs({'randomSeed': int})
    self.randomSeed = randomSeed

  def setArguments(self, args):
    """Define the arguments of the script.

    Alternative to :func:`GaudiApp.setExtraCLIArguments`.

    :param str args: Arguments to pass to the command call
    """
    self._checkArgs({'args': (str,)})
    self.extraCLIArguments = args
    return S_OK()

  def setInputFileFlag(self, inputFileFlag):
    """ Optional: Define inputTypeFlag for k4run input files.

    :param str inputFileFlag: type of event files (LCIO, EventDataSvc)
    """
    self._checkArgs({'inputFileFlag': (str,)})
    self.inputFileFlag = inputFileFlag

  def setOutputFileFlag(self, outputFileFlag):
    """Set the command line parameter to be used for setting the output file

    :param str outputFileFlag: prepend flag to ouput file name (e.g. ``--PodioOutput.filename``)
       Default is ``--out.filename``
    """
    self._checkArgs({'outputFileFlag': (str,)})
    self.outputFileFlag = outputFileFlag

  def setDetectorModelFlag(self, detectorModelFlag):
    """Set the command line parameter to be used for setting the detector model file

    :param str detectorModelFlag: prepend flag to detector model file name (e.g. ``--GeoSvc.detectors``)
       Default is ``--GeoSvc.detectors``
    """
    self._checkArgs({'detectorModelFlag': (str,)})
    self.detectorModelFlag = detectorModelFlag

  def setStartFrom(self, startfrom):
    """Define from where GaudiApp starts to read in the input file.

    :param int startfrom: from where GaudiApp starts to read the input file
    """
    self._checkArgs({'startfrom': int})
    self.startFrom = startfrom

  def setExecutable(self, executableName):
    """Set the executable.

    :param str executableName: Name of the gaudi executable program. Default ``k4Run``
    """
    self._checkArgs({'executableName': (str,)})
    self.executableName = executableName

  def _userjobmodules(self, stepdefinition):
    res1 = self._setApplicationModuleAndParameters(stepdefinition)
    res2 = self._setUserJobFinalization(stepdefinition)
    if not res1["OK"] or not res2["OK"]:
      return S_ERROR('userjobmodules failed')
    return S_OK()

  def _prodjobmodules(self, stepdefinition):
    res1 = self._setApplicationModuleAndParameters(stepdefinition)
    res2 = self._setOutputComputeDataList(stepdefinition)
    if not res1["OK"] or not res2["OK"]:
      return S_ERROR('prodjobmodules failed')
    return S_OK()

  def _checkConsistency(self, job=None):
    """Check consistency of the GaudiApp application.

    Called from the `Job` instance.

    :param job: The instance of the job
    :type job: ~ILCDIRAC.Interfaces.API.NewInterface.Job.Job
    :returns: S_OK/S_ERROR
    """
    parameterName = [pN for pN in job.workflow.parameters.getParametersNames() if 'ConfigPackage' in pN]
    configversion = None
    if parameterName:
      LOG.notice("Found config parameter", parameterName)
      config = job.workflow.parameters.find(parameterName[0])
      configversion = config.value
    # Platform must always be defined
    platform = job.workflow.parameters.find("Platform").value

    if not self.version:
      return S_ERROR('No version found')
    if self.steeringFile:
      if not os.path.exists(self.steeringFile) and not self.steeringFile.lower().startswith("lfn:"):
        res = Exists(self.steeringFile, platform=platform, configversion=configversion)
        if not res['OK']:
          return res

    if self._jobtype != 'User':
      self._listofoutput.append({"outputFile": "@{OutputFile}", "outputPath": "@{OutputPath}",
                                 "outputDataSE": '@{OutputSE}'})

      self.prodparameters['detectorType'] = self.detectortype
      self.prodparameters['slic_detectormodel'] = self.detectorModel

    if not self.startFrom:
      LOG.verbose('No startFrom defined for GaudiApp : start from the beginning')

    return S_OK()

  def _applicationModule(self):

    md1 = self._createModuleDefinition()
    md1.addParameter(Parameter("randomSeed", 0, "int", "", "", False, False,
                               "Random seed for the generator"))
    md1.addParameter(Parameter("executableName", "", "string", "", "", False, False,
                               "Name of the executable"))
    md1.addParameter(Parameter("startFrom", 0, "int", "", "", False, False,
                               "From where GaudiApp starts to read the input file"))
    md1.addParameter(Parameter("detectorModel", "", "string", "", "", False, False,
                               "Detecor model for simulation"))
    md1.addParameter(Parameter("inputFileFlag", "", "string", "", "", False, False,
                               "Flag that define the type flag for input files (lcio, eventdatasvc, ..)"))
    md1.addParameter(Parameter("outputFileFlag", "", "string", "", "", False, False,
                               "Flag that defines the output file flag (e.g. PodioOutput)"))
    md1.addParameter(Parameter("detectorModelFlag", "", "string", "", "", False, False,
                               "Flag that defines the detector model file flag (e.g. GeoSvc)"))
    md1.addParameter(Parameter("debug", False, "bool", "", "", False, False, "debug mode"))
    return md1

  def _applicationModuleValues(self, moduleinstance):

    moduleinstance.setValue("randomSeed", self.randomSeed)
    moduleinstance.setValue("executableName", self.executableName)
    moduleinstance.setValue("startFrom", self.startFrom)
    moduleinstance.setValue("detectorModel", self.detectorModel)
    moduleinstance.setValue("inputFileFlag", self.inputFileFlag)
    moduleinstance.setValue("outputFileFlag", self.outputFileFlag)
    moduleinstance.setValue("detectorModelFlag", self.detectorModelFlag)
    moduleinstance.setValue("debug", self.debug)

  def _checkWorkflowConsistency(self):
    return self._checkRequiredApp()

  def _resolveLinkedStepParameters(self, stepinstance):
    if isinstance(self._linkedidx, six.integer_types):
      self._inputappstep = self._jobsteps[self._linkedidx]
    if self._inputappstep:
      stepinstance.setLink("InputFile", self._inputappstep.getType(), "OutputFile")
    return S_OK()



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Overlay with Marlin and Mokka."""
from __future__ import print_function
from __future__ import absolute_import
from DIRAC.Core.Base import Script
from six.moves import range
Script.parseCommandLine()

from ILCDIRAC.Interfaces.API.NewInterface.UserJob import UserJob
from ILCDIRAC.Interfaces.API.NewInterface.Applications import Mokka, Marlin, OverlayInput
from ILCDIRAC.Interfaces.API.DiracILC import DiracILC

from DIRAC import exit as dexit

d = DiracILC(True, "repo.rep")


n_evts = 500
n_evts_per_job = 100
n_jobs = n_evts / n_evts_per_job

for i in range(n_jobs):
  j = UserJob()

  mo = Mokka()
  mo.setEnergy(3000)
  mo.setVersion("0706P08")
  mo.setSteeringFile("clic_ild_cdr.steer")
  mo.setMacFile("particlegun_electron.mac")
  mo.setOutputFile("MyFile.slcio")
  mo.setNbEvts(n_evts_per_job)
  res = j.append(mo)
  if not res['OK']:
    print(res['Message'])
    break
  ma = Marlin()
  ma.setVersion("v0111Prod")
  ma.setSteeringFile("clic_ild_cdr_steering.xml")
  ma.getInputFromApp(mo)
  ma.setOutputDstFile("mydst_no_ov_%s.slcio" % i)
  res = j.append(ma)
  if not res['OK']:
    print(res['Message'])
    break
  ov = OverlayInput()
  ov.setBXOverlay(60)
  ov.setGGToHadInt(3.2)
  ov.setNbSigEvtsPerJob(n_evts_per_job)
  ov.setBkgEvtType("gghad")
  ov.setDetectorModel("CLIC_ILD_CDR")

  res = j.append(ov)
  if not res['OK']:
    print(res['Message'])
    break
  ma2 = Marlin()
  ma2.setVersion("v0111Prod")
  ma2.setSteeringFile("clic_ild_cdr_steering_overlay.xml")
  ma2.getInputFromApp(mo)
  ma2.setOutputDstFile("mydst_ov_%s.slcio" % i)
  res = j.append(ma2)
  if not res['OK']:
    print(res['Message'])
    break
  j.setOutputSandbox(["mydst_no_ov_%s.slcio" % i, "mydst_ov_%s.slcio" % i, "*.log"])
  j.setName("SingleElectron_%s" % i)
  j.setJobGroup("singleElectrons")

  j.submit(d)

dexit(0)



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Tests for ILCTransformationSystem.Utilities.Utilities."""
from __future__ import absolute_import
from ILCDIRAC.ILCTransformationSystem.Utilities.Utilities import Task


def test_Constructor():
  """Test for the constructor and functions used inside."""
  metaInput = {'ProdID': 123, 'Energy': '100'}
  parameterDict = {'process': 'higgses'}
  eventsPerJob = 22
  theTask = Task(metaInput, parameterDict, eventsPerJob)
  assert theTask.meta['ProdID'] == 123
  assert theTask.getProdName() == 'higgses_100'

  metaInput = {'ProdID': 124, 'Energy': '100'}
  metaPrev = {'ProdID': 123, 'Energy': '100'}
  theTask = Task(metaInput, parameterDict, eventsPerJob, metaPrev)
  assert theTask.meta['ProdID'] == 124
  assert theTask.meta['NumberOfEvents'] == eventsPerJob
  theTask.taskName = 'aTask'
  assert theTask.getProdName() == 'higgses_100_aTask'
  theTask.sourceName = 'theSource'
  assert theTask.getProdName() == 'higgses_100_theSource_aTask'

  theTask = Task(metaInput, parameterDict, eventsPerJob, metaPrev)
  assert theTask.meta['ProdID'] == 124
  assert theTask.meta['NumberOfEvents'] == eventsPerJob

  theTask = Task(metaInput, parameterDict, None, metaPrev)
  assert theTask.meta['ProdID'] == 124
  assert theTask.meta.get('NumberOfEvents') is None

  theTask = Task(metaInput, parameterDict, eventsPerJob=0, metaPrev=metaPrev)
  assert theTask.meta.get('NumberOfEvents') == 0
  assert theTask.meta.get('ProdID') == 124

  # expect eventsPerBaseFile
  theTask = Task(metaInput, parameterDict, eventsPerJob=0, metaPrev=metaPrev, eventsPerBaseFile=333)
  assert theTask.meta.get('NumberOfEvents') == 333


def test_repr():
  """Test the str and repr functions for Task."""
  task = Task({}, {}, 1)
  expectation = '{' + ',\n'.join('%r:%r' % (p, val) for p, val in sorted(vars(task).items())) + '}'
  assert repr(task).replace(' ', '') == expectation.replace(' ', '')



#!/usr/local/env python
#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Test GetSRMFile  module."""

from __future__ import absolute_import
import unittest
from mock import patch, MagicMock as Mock

from DIRAC import S_OK, S_ERROR
from ILCDIRAC.Interfaces.API.NewInterface.Applications import GetSRMFile
from Tests.Utilities.GeneralUtils import assertEqualsImproved, assertDiracFailsWith, \
    assertDiracSucceeds

__RCSID__ = "$Id$"

MODULE_NAME = 'ILCDIRAC.Interfaces.API.NewInterface.Applications.GetSRMFile'

# pylint: disable=protected-access


class GetSRMFileTestCase(unittest.TestCase):
  """Base class for the GetSRMFile test cases."""

  def setUp(self):
    """set up the objects."""
    self.gsf = GetSRMFile({})

  def test_setfiles(self):
    self.assertFalse(self.gsf._errorDict)
    self.gsf.setFiles({'mydict': True, 'soimething': 138, True: []})
    self.assertFalse(self.gsf._errorDict)
    assertEqualsImproved(self.gsf.files, {'mydict': True, 'soimething': 138, True: []}, self)
    self.gsf.setFiles([{'some_dict': True}, {}, {True: 184, False: [], 'bla': 'aiejf'}])
    self.assertFalse(self.gsf._errorDict)
    assertEqualsImproved(self.gsf.files, [{'some_dict': True}, {},
                                            {True: 184, False: [], 'bla': 'aiejf'}], self)

  def test_setfiles_wrongtype(self):
    self.assertFalse(self.gsf._errorDict)
    self.gsf.setFiles(1498)
    assertEqualsImproved(len(self.gsf._errorDict['setFiles']), 1, self)

  def test_userjobmodules(self):
    module_mock = Mock()
    assertDiracSucceeds(self.gsf._userjobmodules(module_mock), self)

  def test_prodjobmodules(self):
    assertDiracFailsWith(self.gsf._prodjobmodules(Mock()),
                          'should not use in production', self)

  def test_userjobmodules_fails(self):
    with patch('%s._setUserJobFinalization' % MODULE_NAME, new=Mock(return_value=S_OK('something'))),\
        patch('%s._setApplicationModuleAndParameters' % MODULE_NAME, new=Mock(return_value=S_ERROR('some_test_err'))):
      assertDiracFailsWith(self.gsf._userjobmodules(None),
                            'userjobmodules method failed', self)

  def test_checkconsistency_nofiles(self):
    assertDiracFailsWith(self.gsf._checkConsistency(), 'file list was not defined', self)

  def test_checkconsistency(self):
    self.gsf.setFiles({'file': '/invalid/dir/mycoolfile.txt', 'something': True})
    assertDiracSucceeds(self.gsf._checkConsistency(), self)
    assertEqualsImproved(self.gsf.outputFile, 'mycoolfile.txt', self)

  def test_checkconsistency_list(self):
    self.gsf.setFiles([{'file': '/invalid/dir/mycoolfile.txt', 'something': True},
                         {'file': '/some/other/dir/other_file.123.stdhep'}])
    assertDiracSucceeds(self.gsf._checkConsistency(), self)
    assertEqualsImproved(self.gsf.outputFile, 'mycoolfile.txt;other_file.123.stdhep', self)

  def test_addparams(self):
    assertDiracSucceeds(self.gsf._addParametersToStep(Mock()), self)

  def test_addparams_fails(self):
    with patch.object(self.gsf, '_addBaseParameters', new=Mock(return_value=S_ERROR('my_patched_testerr'))):
      assertDiracFailsWith(self.gsf._addParametersToStep(None),
                            'failed to set base parameters', self)



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Unit tests for the CalibrationAgent."""

from __future__ import absolute_import
import filecmp
import os
import unittest

from ILCDIRAC.CalibrationSystem.Utilities.fileutils import binaryFileToString, stringToBinaryFile


__RCSID__ = "$Id$"

MODULE_NAME = 'ILCDIRAC.CalibrationSystem.Agent.CalibrationAgent'


class TestsFileUtils(unittest.TestCase):
  """Test the utilities for the CalibrationSystem."""

  def setUp(self):
    """Set up the objects."""
    self.targetFile = "targetFile.root"

  def tearDown(self):
    """Tear down the objects."""
    # os.unlink( self.targetFile )
    pass

  def test_binaryToString(self):
    """Test stringToBinaryFile function."""
    filename = os.path.join(os.environ['ILCDIRAC_BASE_FOLDER'], 'Tests', 'Files', 'input.root')
    content = binaryFileToString(filename)
    stringToBinaryFile(content, self.targetFile)
    self.assertTrue(filecmp.cmp(filename, self.targetFile), "Files are not the same any more")



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""The calibration client offers the interface for user and calibration worker nodes to talk to the calibration service.

The user use this interface to monitor and control calibrations.
The worker nodes use this interface to ask for new parameters, their event slices and inform the service
about the results of their reconstruction.
"""

from __future__ import absolute_import
from DIRAC.Core.Base.Client import Client
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from ILCDIRAC.CalibrationSystem.Utilities.fileutils import binaryFileToString
from ILCDIRAC.CalibrationSystem.Service.CalibrationRun import CalibrationPhase

__RCSID__ = "$Id$"
LOG = gLogger.getSubLogger(__name__)


class CalibrationClient(Client):
  """Provide an interface for user and worker nodes to talk to Calibration service.

  Contains interfaces to fetch the necessary data from the service to worker node and to report the results back.
  """

  def __init__(self, calibrationID=None, workerID=None, **kwargs):
    """Initialize the client.

    :param int workerID: ID of this worker
    :param int calibrationID: ID of the calibration run this worker belongs to
    :returns: None
    """
    super(CalibrationClient, self).__init__(**kwargs)
    self.setServer('Calibration/Calibration')
    self.calibrationID = calibrationID
    self.workerID = workerID
    self.currentStep = -1  # initial parameter request
    self.currentPhase = CalibrationPhase.ECalDigi
    self.currentStage = 1
    self.parameterSet = None
    self.log = LOG
    self.ops = Operations()
    self.maximumReportTries = self.ops.getValue('Calibration/MaximumReportTries', 10)  # TODO add this to CS

  def getInputDataDict(self, calibrationID=None, workerID=None):
    """Get input data dict. If no arguments are passed use ids which belong to the current class instance.

    :param int calibrationID: ID of the calibration
    :param int workerID: ID of this worker
    :returns: S_OK or S_ERROR
    """
    if calibrationID is None and self.calibrationID is None:
      return S_ERROR("Specify calibrationID")
    if workerID is None and self.workerID is None:
      return S_ERROR("Specify workerID")

    calibIDToUse = calibrationID if calibrationID is not None else self.calibrationID
    workerIDToUse = workerID if workerID is not None else self.workerID

    return self._getRPC().getInputDataDict(calibIDToUse, workerIDToUse)

  def requestNewParameters(self):
    """Fetch new parameter set from the service and updates the step counter in this object with the new value.

    Throws a ValueError if the calibration ended already.
    :returns: dict with 4 keys: calibrationIsFinished (bool), parameters (dict), currentPhase (int), currentStage (int)
    or None if no new parameters are available yet
    :rtype: list
    """
    res = self.getNewParameters(self.calibrationID, self.currentStep)
    if res['OK']:
      returnValue = res['Value']
      # FIXME calibrationRun state will be updated number of worker times while only one time is enough
      if returnValue is not None:
        self.currentPhase = returnValue['currentPhase']
        self.currentStage = returnValue['currentStage']
        self.currentStep = returnValue['currentStep']
    return res

  def requestNewPhotonLikelihood(self):
    """Get new photon likelihood file."""
    res = self.getNewPhotonLikelihood(self.calibrationID)
    if res['OK']:
      return res['Value']
    else:
      return None

  def reportResult(self, outFileName):
    """Send result of calibration step at the node (.root or .xml file) to the service.

    Send the root file from PfoAnalysis or PandoraLikelihoodDataPhotonTraining.xml from photon training
    back to the service procedure

    :param outFileName: Output file from one calibration iteration
    :returns: None
    """
    attempt = 0
    resultString = binaryFileToString(outFileName)

    while attempt < self.maximumReportTries:
      res = self.submitResult(self.calibrationID, self.currentStage, self.currentPhase,
                                                 self.currentStep, self.workerID, resultString)
      if res['OK']:
        return S_OK()
      self.log.warn("Failed to submit result, try", "%s: %s " % (attempt, res['Message']))
      attempt += 1

    return S_ERROR('Could not report result back to CalibrationService.')

  def createCalibration(self, inputFiles, numberOfEventsPerFile, calibSettingsDict):
    """Create calibration run (series of calibration iterations).

    :param dict inputFiles: Input files for the calibration. Dictionary.
                            Mandatory keys: 'zuds', 'kaon', 'muon', 'gamma'.
    :param dict numberOfEventsPerFile: Number of events per type of input file. Dictionary.
                                       Mandatory keys: 'zuds', 'kaon', 'muon', 'gamma'.
    :param dict calibSettingsDict: Calibration configuration settings reqired to start calibration. Dictionary.
    :returns: S_OK containing ID of the calibration, used to retrieve results etc
    :rtype: dict
    """
    res = self._getRPC().createCalibration(inputFiles, numberOfEventsPerFile, calibSettingsDict)
    if not res['OK']:
      LOG.error('Error while executing createCalibration! Error message: ', res['Message'])
    return res
